# -*- coding: utf-8 -*-
"""
Created on Mon Jun 17 13:05:37 2019

@author: Jose L. L. Elvira
"""

"""Loads data files exported from the Vicon Nexus"""

# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "v4.2.0"
__date__ = "25/03/2025"

"""
Updates:
    25/03/2025, v4.2.0
        - Incuded general function to distribute according to "engine". 

    10/03/2025, v4.1.2
        - Adapted to biomdp with translations. 

    06/03/2025, v.4.1.1
        - Included type specifications in functions.
    
    25/11/2023, v.4.1.0
        - La función read_vicon_csv_pl admite el parámetro to_dataarray 
          para decidir si se devuelve un xr.DataArray o un polars dataframe.
    
    07/05/2023, v.4.0.4
        - Mejorada lectura con polars
    
    22/04/2023, v.4.0.3
        - Corregido que n Polars actual (0.17.2) no coindicen los saltos de
          línea cuando empieza con línea en blanco.
        - Optimizado código de lectura con Polars.
          
    17/04/2023, v.4.0.2
        - Corregido error nombre variable archivo en read_vicon_csv_pl.
        - Optimizada parte lectura con open en versión Polars.
        - Puede cargar por separado datos EMG o Forces, aunque vayan en el
          mismo archivo.
        - Cambiada nombre variables a n_var por conflicto en xarray, .var
          calcula la varianza.
        - En función con Polars, La selección de variables la hace en el df
          Polars, no en el dataarray.

    27/03/2023, v.4.0.0
        - Incluida una versión que lee con Polars y lo pasa a DataArray
          directamente (read_vicon_csv_pl), mucho más rápida.
        - La versión que lee con Polars es capaz de leer csv con encabezados
          repetidos.
          
    07/06/2022, v.3.0.0
        - Intento de corrección cuando tiene que leer csv con variables EMG
          modeladas, que se interfieren entre las xyz.
        
    09/04/2022, v.2.3.0
        - Habilitado para cargar como dataframe y dataArray EMG Noraxon en modo Devices.
        - Incluye el tiempo en una columna.

    29/03/2021, v2.1.1
        - Incluido parámetro 'header_format' para que devuelva el encabezado como 'flat' en una sola línea (variable_x, variable_y, ...) o en dos líneas ((variable,x), (variable,y), ...).
            
    28/03/2021, v2.1.1
        - Mejorada lectura con Pandas. Ahora puede cargar archivos que empiezan sin datos en las primeras líneas.

	21/03/2021, v2.1.0
        - Cambiado lector del bloque de archivos por pd.read_csv con número de columnas delimitado a los que carga en las variables (quitando los de velocidad y aceleración)
        - Solucionado fallo al leer frecuencia cuando terminaba la línea rellenando con separadores (como al exportar en Excel)

    10/01/2021, v2.0.1
        - Ajustado para que pueda devolver xArray con Model Outputs

    13/12/2020, v2.0.0
        - Con el argumento to_dataarray se puede pedir que devuelva los datos en formato xArray
"""
from typing import List, Any
import numpy as np

import pandas as pd
import xarray as xr

from pathlib import Path


# import scipy.signal

import os


# =============================================================================
# %% Functions
# =============================================================================


def read_vicon_csv(
    file: str | Path,
    section: str = "Model Outputs",
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
    sep: str = ",",
    engine: str = "polars",
    raw: bool = False,
) -> Any:
    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    # Ensures that file is a csv file
    file = file.with_suffix(".csv")

    if engine == "polars":
        da = read_vicon_csv_pl(
            file,
            section=section,
            n_vars_load=n_vars_load,
            coincidence=coincidence,
            sep=sep,
            raw=raw,
        )
    elif engine == "polars2":
        da = read_vicon_csv_pl2(
            file,
            section=section,
            n_vars_load=n_vars_load,
            coincidence=coincidence,
            sep=sep,
            raw=raw,
        )
    elif engine == "pandas":
        da = read_vicon_csv_pd(
            file,
            section=section,
            sep=sep,
            n_vars_load=n_vars_load,
            raw=True,
        )
    else:
        raise ValueError(f"Engine {engine} not valid\nTry with 'polars' or 'pandas'")

    return da


def read_vicon_csv_pl(
    file: str | Path,
    section: str = "Model Outputs",
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
    sep: str = ",",
    raw: bool = False,
):  # -> xr.DataArray | pl.DataFrame:
    """
    Reads data from a Vicon csv file and returns a Polars DataFrame or xarray DataArray

    Parameters
    ----------
    file : string or path of the file
        DESCRIPTION.
    section : string, optional
        Kind of data variables to load.
        Options: 'Trajectories', 'Model Outputs', 'Forces', 'EMG', 'Model Outputs EMG'
        The default is 'Model Outputs'.
    n_vars_load : list, optional
        DESCRIPTION. The default is None.
    coincidence: string
        When selecting which variables to load, allows strings containing the
        indicated or forces to be exact.
        Options: 'similar', 'exact'. The default is 'similar'.
    sep : string, optional
        Separator used in the csv file. The default is ','.
    to_dataarray : bool, optional
        Transforms data to dataarray, otherway is polars dataframe. The default is True.

    Returns
    -------
    Xarray DataArray.

    """

    try:
        import polars as pl
    except:
        raise ImportError(
            "Polars package not instaled. Install it if you want to use the accelerated version"
        )

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    if section in ["Forces", "EMG"]:
        n_block = "Devices"
    elif section == "Model Outputs EMG":
        n_block = "Model Outputs"
    elif section in ["Trajectories", "Model Outputs"]:
        n_block = section
    else:
        raise ValueError(
            f"Section '{section}' not recognized\nTry with 'full', 'Trajectories', 'Model Outputs', 'Forces', 'EMG', 'Model Outputs EMG'"
        )

    # ----Check for blank lines. In current Polars (0.17.2) line breaks do not match when starting with blank line
    # Vicon Nexus writes a first empty line when export Events is selected but thre are no events
    with open(file, mode="rt") as f:
        offset_blank_ini = 0
        while f.readline() in ["\r\n", "\n", "ï»¿\n", r"\n", "\ufeff"]:
            # print("blank line")
            offset_blank_ini += 1
    # Second Check for blank lines (some times the previous one does not work)
    if offset_blank_ini == 0:
        with open(file, mode="rt") as f:
            line = f.readline()
            if not any(
                filter(
                    lambda x: x in line,
                    ["Events", "Trajectories", "Model Outputs", "Devices"],
                )
            ):
                offset_blank_ini += 1

    # ----Search section position and length
    ini_section = None
    end_section = None
    freq = 1
    with open(file, mode="rt") as f:
        num_lin = 0

        # Scrolls through the entire file to find the start and end of the section and the number of lines
        for line in f:
            # Search for section start label
            if ini_section is None and n_block in line:
                ini_section = num_lin
                # The frequency is below the section label
                freq = int(
                    f.readline().replace(sep, "")
                )  # removes the separator for cases where the file has been saved with Excel (full line with separator)

                # #Load columns names
                # n_head = str(f.readline()[:-1]).split(sep) #variable names
                # n_subhead = str(f.readline()[:-1]).lower().split(sep) #coordinate names (x,y,z)
                # num_lin+=3
                num_lin += 1
            # When start found, search the end
            if ini_section is not None and end_section is None and line == "\n":
                end_section = num_lin - 1
                break

            num_lin += 1

        if end_section is None:
            end_section = num_lin - 1
        # file_end = num_lin

    if ini_section is None:
        raise Exception(f"Section header '{section}' not found")

    # ----Load data from file
    # pl.read_csv(file, truncate_ragged_lines=True)
    df = pl.read_csv(
        file,
        has_header=True,
        skip_rows=ini_section + 2 - offset_blank_ini,
        n_rows=end_section - ini_section - 2,
        truncate_ragged_lines=True,
        separator=sep,
        # columns=range(len(n_vars_merged)),
        # new_columns=n_vars_merged, separator=sep
    )

    n_head = df.columns
    n_subhead = list(np.char.lower(df.slice(0, 1).to_numpy()[0].astype(str)))

    # Remove subhead and units rows
    df = df.slice(2, None)

    if n_block in ["Trajectories", "Model Outputs", "Model Outputs EMG"]:
        n_vars_merged = ["Frame", "Sub Frame"]
        for i in range(2, len(n_subhead)):
            if (
                n_subhead[i] in "x"
                and "'" not in n_subhead[i]
                and "''" not in n_subhead[i]
            ):
                # print(n_subhead[i], n_head[i])
                n_vars_merged.append(n_head[i].split(":")[1] + "_" + n_subhead[i])  # X
                n_vars_merged.append(
                    n_head[i].split(":")[1] + "_" + n_subhead[i + 1]
                )  # Y
                n_vars_merged.append(
                    n_head[i].split(":")[1] + "_" + n_subhead[i + 2]
                )  # Z
            elif "emg" in n_subhead[i]:
                # print(n_subhead[i], n_head[i])
                n_vars_merged.append(n_head[i].split(":")[1] + "_" + n_subhead[i])

        # Rename headers
        df = df.rename(dict(zip(n_head, n_vars_merged)))

        if section == "Model Outputs EMG":
            df = df.select(pl.col("^*_emg.*$"))
            df = (
                df.rename(dict(zip(df.columns, [n[:-3] for n in df.columns]))).select(
                    pl.exclude("^*_duplicate.*$")
                )  # remove duplicates
                # .slice(2, None) #remove subhead and units rows
            )
        else:  # Trajectories and Model Outputs
            df = (
                df.select(pl.exclude(["Frame", "Sub Frame"]))
                .select(pl.exclude("^*_duplicate.*$"))  # remove duplicates
                .select(pl.exclude("^*_emg.*$"))  # remove 1D EMG variables
                # .slice(2, None) #remove subhead and units rows
            )

    elif n_block == "Devices":
        if section == "EMG":
            # n_head = n_head[:len(n_subhead)] #ajusta el número de variables
            # n_vars_merged = rename_duplicates(n_head)
            # selection = [s for s in n_vars_merged[2:] if 'EMG' in s]
            # selection2 = selection

            df = (
                df.select(pl.col("^*EMG.*$")).select(
                    pl.exclude("^*_duplicate.*$")
                )  # remove duplicates
                # .slice(2, None) #remove subhead and units rows
            )

        elif section == "Forces":
            n_vars_merged = ["Frame", "Sub Frame"]
            for i in range(2, len(n_subhead)):
                if (
                    "x" in n_subhead[i]
                    and "'" not in n_subhead[i]
                    and "''" not in n_subhead[i]
                ):
                    # print(n_subhead[i], n_head[i])
                    n_vars_merged.append(n_head[i] + "_" + n_subhead[i])  # X
                    n_vars_merged.append(n_head[i] + "_" + n_subhead[i + 1])  # Y
                    n_vars_merged.append(n_head[i] + "_" + n_subhead[i + 2])  # Z
                elif "v" in n_subhead[i]:
                    n_vars_merged.append(n_head[i] + "_" + n_subhead[i])

            df = (
                df.rename(dict(zip(df.columns, n_vars_merged)))
                .select(pl.exclude("^*EMG.*$"))
                .select(pl.exclude("^*_duplicate.*$"))  # remove duplicates
                # .slice(2, None) #remove subhead and units rows
            )

    # ----Filter variables
    if n_vars_load is not None:
        if not isinstance(n_vars_load, list):
            n_vars_load = [n_vars_load]  # in case a string is passed
        if coincidence == "similar":
            selection = [s for s in df.columns if any(xs in s for xs in n_vars_load)]
        elif coincidence == "exact":
            selection = n_vars_load
        else:
            raise ValueError(
                f"Coincidence '{coincidence}' not recognized\nTry with 'similar' or 'exact'"
            )
        df = df.select(pl.col(selection))

    if raw:  # keep Polars dataframe format
        # Add time column
        df.insert_column(0, pl.lit(np.arange(len(df)) / freq).alias("time"))
        df.insert_column(0, pl.lit(file.stem).alias("ID"))

        # Add section as prefix
        df = df.select(pl.all().name.prefix(f"{section}_"))

        # df = df.with_columns(
        #     pl.lit(np.arange(len(df)) / freq).alias("time"),
        #     # pl.all()
        # )
        return df

    else:  # ----Transform polars to xarray
        if section in ["Trajectories", "Model Outputs", "Forces"]:
            # Decompose on its axes
            x = df.select(pl.col("^*x$")).to_numpy()
            y = df.select(pl.col("^*y$")).to_numpy()
            z = df.select(pl.col("^*z$")).to_numpy()
            data = np.stack([x, y, z])

            ending = -3 if section == "Forces" else -2
            coords = {
                "axis": ["x", "y", "z"],
                "time": np.arange(data.shape[1]) / freq,
                "n_var": [x[:ending] for x in df.columns if "x" in x[-1]],
            }
            da = (
                xr.DataArray(
                    data=data,
                    dims=coords.keys(),
                    coords=coords,
                )
                .astype(float)
                .transpose("n_var", "axis", "time")
            )

        elif section in ["EMG", "Model Outputs EMG"]:
            data = df.to_numpy().T
            coords = {
                "n_var": df.columns,  # selection2,
                "time": np.arange(df.shape[0]) / freq,
            }
            da = xr.DataArray(
                data=data,
                dims=coords.keys(),
                coords=coords,
            ).astype(float)
        else:
            raise ValueError(
                f"Section '{section}' not recognized\nTry with 'full', 'Trajectories', 'Model Outputs', 'Forces', 'EMG', 'Model Outputs EMG'"
            )
        da.name = section
        da.attrs["freq"] = freq
        da.time.attrs["units"] = "s"
        if section == "Trajectories":
            da.attrs["units"] = "mm"
        elif "EMG" in section:
            da.attrs["units"] = "V"
        elif "Forces" in section:
            da.attrs["units"] = "N"

        return da


def rename_duplicates(names: List[str]):
    cols = pd.Series(names)
    if len(names) != len(cols.unique()):
        for dup in cols[cols.duplicated()].unique():
            cols[cols[cols == dup].index.values.tolist()] = [
                dup + "." + str(i) if i != 0 else dup for i in range(sum(cols == dup))
            ]
        names = list(cols)
    return names


# -----------------------------------------------------------------------------
# EXPÈRIMENTAL
def read_vicon_csv_pl2(
    file: str | Path,
    section: str = "Model Outputs",
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
    sep: str = ",",
    raw: bool = False,
) -> xr.DataArray:
    """
    Reads data from a Vicon csv file and returns a Polars DataFrame or xarray DataArray

    Parameters
    ----------
    file : string or path of the file
        DESCRIPTION.
    section : string, optional
        Kind of data variables to load.
        Options: 'Trajectories', 'Model Outputs', 'Forces', 'EMG', 'Model Outputs EMG'
        The default is 'Model Outputs'.
    n_vars_load : list, optional
        DESCRIPTION. The default is None.
    coincidence: string
        When selecting which variables to load, allows strings containing the
        indicated or forces to be exact.
        Options: 'similar', 'exact'. The default is 'similar'.
    sep : string, optional
        Separator used in the csv file. The default is ','.
    to_dataarray : bool, optional
        Transforms data to dataarray, otherway is polars dataframe. The default is True.

    Returns
    -------
    Xarray DataArray.

    """
    try:
        import polars as pl
    except:
        raise ImportError(
            "Polars package not instaled. Install it if you want to use the accelerated version"
        )

    print("alternativo")
    if section in ["Forces", "EMG"]:
        n_block = "Devices"
    elif section == "Model Outputs EMG":
        n_block = "Model Outputs"
    elif section in ["Trajectories", "Model Outputs"]:
        n_block = section
    else:
        raise ValueError(
            f"Section '{section}' not recognized\nTry with 'Trajectories', 'Model Outputs', 'Forces', 'EMG', 'Model Outputs EMG'"
        )

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    # Ensures that file is a csv file
    file = file.with_suffix(".csv")

    n_head = []
    n_subhead = []

    # ----Check for blank lines. In current Polars (0.17.2) line breaks do not match when starting with blank line
    offset_blank_ini = 0
    with open(file, mode="rt") as f:
        while f.readline() in ["\r\n", "\n", "ï»¿\n", r"\n", "\ufeff"]:
            # print("blank line")
            offset_blank_ini += 1
    # Second Check for blank lines (some times the previous one does not work)
    if offset_blank_ini == 0:
        with open(file, mode="rt") as f:
            line = f.readline()
            # print(line)
            if not any(
                filter(
                    lambda x: x in line,
                    ["Events", "Trajectories", "Model Outputs", "Devices"],
                )
            ):
                offset_blank_ini += 1

    # ----Search section position and length
    ini_section = None
    end_section = None
    with open(file, mode="rt") as f:
        num_lin = 0

        # Scrolls through the entire file to find the start and end of the section and the number of lines
        for line in f:
            # Search for section start label
            if ini_section is None and n_block in line:
                ini_section = num_lin
                # The frequency is below the section label
                freq = int(
                    f.readline().replace(sep, "")
                )  # removes the separator for cases where the file has been saved with Excel (full line with separator)

                # Load columns names
                n_head = str(f.readline()[:-1]).split(sep)  # variable names
                n_subhead = (
                    str(f.readline()[:-1]).lower().split(sep)
                )  # coordinate names (x,y,z)
                num_lin += 3

            # When start found, search the end
            if ini_section is not None and end_section is None and line == "\n":
                end_section = num_lin - 1
                break
            num_lin += 1
        # file_end = num_lin

    if ini_section is None:
        raise Exception("Section header not found")

    # ----Assign header labels
    if n_block == "Devices":
        if section == "EMG":
            n_head = n_head[: len(n_subhead)]  # adjust number of variables
            n_vars_merged = rename_duplicates(n_head)
            selection = [s for s in n_vars_merged[2:] if "EMG" in s]
            selection2 = selection

        elif section == "Forces":
            n_vars_merged = [h + "_" + sh for h, sh in zip(n_head, n_subhead)]
            n_vars_merged = rename_duplicates(n_vars_merged)
            selection = [s for s in n_vars_merged[2:] if "EMG" not in s]
            selection2 = []
            for i in range(2, len(n_subhead)):
                if "x" in n_subhead[i] and "EMG" not in n_subhead[i]:
                    # print(n_subhead[i], n_head[i])
                    selection2.append(n_head[i] + "_" + n_subhead[i][-1])  # x
                    selection2.append(n_head[i] + "_" + n_subhead[i + 1][-1])  # y
                    selection2.append(n_head[i] + "_" + n_subhead[i + 2][-1])  # z

    elif n_block in ["Trajectories", "Model Outputs", "Model Outputs EMG"]:
        n_vars_merged = ["Frame", "Sub Frame"]
        for i in range(2, len(n_subhead)):
            if (
                n_subhead[i] in "xX"
                and "'" not in n_subhead[i]
                and "''" not in n_subhead[i]
            ):
                # print(n_subhead[i], n_head[i])
                n_vars_merged.append(n_head[i].split(":")[1] + "_" + n_subhead[i])  # X
                n_vars_merged.append(
                    n_head[i].split(":")[1] + "_" + n_subhead[i + 1]
                )  # Y
                n_vars_merged.append(
                    n_head[i].split(":")[1] + "_" + n_subhead[i + 2]
                )  # Z
            elif "emg" in n_subhead[i]:
                # print(n_subhead[i], n_head[i])
                n_vars_merged.append(n_head[i].split(":")[1] + "_" + n_subhead[i])
        selection = set(n_vars_merged[2:])  # remove duplicates
        if section == "Model Outputs EMG":
            selection = sorted([s for s in selection if "_emg" in s])
            selection2 = [s[:-4] for s in selection]
        else:
            selection = sorted([s for s in selection if "_emg" not in s])
            selection2 = selection
        n_vars_merged = rename_duplicates(n_vars_merged)

    # ----Load data from file
    df = (
        pl.read_csv(
            file,
            has_header=False,
            skip_rows=ini_section + 3 - offset_blank_ini,
            n_rows=end_section - ini_section - 2,
            columns=range(len(n_vars_merged)),
            new_columns=n_vars_merged,
            separator=sep,
        )
        .select(pl.col(selection))
        .rename(dict(zip(selection, selection2)))
        .slice(2, None)  # remove subhead and units rows
    )

    # ----Filter variables
    if n_vars_load:
        if not isinstance(n_vars_load, list):
            n_vars_load = [n_vars_load]  # in case a string is passed
        if coincidence == "similar":
            selection2 = [s for s in selection2 if any(xs in s for xs in n_vars_load)]
        elif coincidence == "exact":
            selection2 = n_vars_load
        df = df.select(pl.col(selection2))

    if raw:  # keep Polars dataframe format
        # Add time column
        df = df.with_columns(
            pl.lit(np.arange(len(df)) / freq).alias("time"),
            # pl.all()
        )
        return df

    else:  # ----Transform polars to xarray
        if section in ["EMG", "Model Outputs EMG"]:
            data = df.to_numpy().T
            coords = {
                "n_var": selection2,
                "time": np.arange(df.shape[0]) / freq,
            }
            da = xr.DataArray(
                data=data,
                dims=coords.keys(),
                coords=coords,
            ).astype(float)

        elif section in ["Trajectories", "Model Outputs", "Forces"]:
            # Decompose on its axes. Those that end in the coordinate or if they are repeated
            x = df.select(pl.col("^*_x|_x.$")).to_numpy()
            y = df.select(pl.col("^*_y|_y.$")).to_numpy()
            z = df.select(pl.col("^*_z|_z.*$")).to_numpy()
            data = np.stack([x, y, z])

            ending = -3 if section == "Forces" else -2
            coords = {
                "axis": ["x", "y", "z"],
                "time": np.arange(data.shape[1]) / freq,
                "n_var": [x[:ending] for x in df.columns if "_x" in x or "_X" in x],
            }
            da = (
                xr.DataArray(
                    data=data,
                    dims=coords.keys(),
                    coords=coords,
                )
                .astype(float)
                .transpose("n_var", "axis", "time")
            )
        else:
            raise ValueError(f"Section {section} not recognized")

        da.name = section
        da.attrs["freq"] = freq
        da.time.attrs["units"] = "s"
        if section == "Trajectories":
            da.attrs["units"] = "mm"
        elif "EMG" in section:
            da.attrs["units"] = "V"
        elif "Forces" in section:
            da.attrs["units"] = "N"
        return da


# -----------------------------------------------------------------------------


def read_vicon_csv_pd(
    file: str | Path,
    section: str = "Model Outputs",
    sep: str = ",",
    return_freq: bool = False,
    to_dataarray: bool = False,
    n_vars_load: List[str] | None = None,
    header_format: str = "flat",
    raw: bool = False,
):
    """
    Reads data from a Vicon csv file and returns a Polars DataFrame or xarray DataArray

    Parameters
    ----------
    file : str or Path
        Path to the csv file.
    section : str, optional
        Section to be read. Options are 'Trajectories', 'Model Outputs', 'Forces', 'EMG'.
        Defaults to 'Model Outputs'.
    sep : str, optional
        Separator used in the csv file. Defaults to ','.
    return_freq : bool, optional
        If True, it returns the sampling frequency of the data. Defaults to False.
    to_dataarray : bool, optional
        If True, it returns a xarray DataArray. Defaults to False.
    header_format : str, optional
        Format of the header. Options are 'flat' and 'hierarchical'. Defaults to 'flat'.

    Returns
    -------
    df : Polars DataFrame
        Data with columns as variables and rows as time.
    da : xarray DataArray
        Data with dimensions (n_var, axis, time). Only returned if `to_dataarray` is True.
    freq : float
        Sampling frequency of the data. Only returned if `return_freq` is True.

    Examples
    --------
    >>> dfData = read_vicon_csv_pd(file_name, section='Model Outputs')
    >>> dfData, frecuencia = read_vicon_csv_pd(file_name, section='Trajectories', return_freq=True)

    >>> #Con formato dataarray de xArray
    >>> daDatos = read_vicon_csv_pd(file_name, section='Trajectories', to_dataarray=True)

    """

    try:
        import pandas as pd
    except:
        raise ImportError("Pandas package not instaled")

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    with open(file, mode="rt") as f:
        num_line = 0
        # Search section label
        line = f.readline()
        while section not in line:
            if line == "":
                raise Exception(f"Section header '{section}' not found")

            num_line += 1
            line = f.readline()

        ini_section = num_line

        # Frequency after the section label
        line = f.readline()
        freq = int(
            line.replace(sep, "")
        )  # quita el separador para los casos en los que el archivo ha sido guardado con Excel (completa línea con separador)

        # Load column names
        # line = f.readline()
        nomColsVar = str(f.readline()[:-1]).split(sep)  # nombreVariables
        nomCols = str(f.readline()[:-1]).split(sep)  # name coordinates X,Y,Z.
        # nomCols = [s.lower() for s in nomCols]

        # Finds the end of the section
        while line != "\n":
            if line == "":
                raise Exception(f"End of section '{section}' not found")

            num_line += 1
            # print('line '+ str(num_line))
            line = f.readline()

    end_section = num_line - 1  # removes 1 to compensate the empty line

    # Counts total line number
    with open(file, mode="rt") as f:
        file_end = len(f.readlines())

    # Column labels
    if section == "Devices":
        n_vars = ["Frame", "Sub Frame"] + nomColsVar[
            2:-1
        ]  # removes the last, which is ''
        # nomVars2=list(filter(lambda c: c!='', n_vars))

    else:  # Trajectories and Models
        # primero asigna los nombres según el propio archivo
        # n_vars=['Frame', 'Sub Frame']
        # for i in range(2,len(nomCols),3):
        #     if "'" not in nomCols[i] and "''" not in nomCols[i] and 'EMG' not in nomCols[i]: #elimina las posibles columnas de velocidad y aceleración
        #         print(nomCols[i])
        #         n_vars.append(nomColsVar[i].split(':')[1]+'_' + nomCols[i])#X
        #         n_vars.append(nomColsVar[i].split(':')[1]+'_' + nomCols[i+1])#Y
        #         n_vars.append(nomColsVar[i].split(':')[1]+'_' + nomCols[i+2])#Z

        n_vars = ["Frame", "Sub Frame"]
        for i in range(2, len(nomCols)):
            if nomCols[i] in "xX" and "'" not in nomCols[i] and "''" not in nomCols[i]:
                # print(nomCols[i], nomColsVar[i])
                n_vars.append(nomColsVar[i].split(":")[1] + "_" + nomCols[i])  # X
                n_vars.append(nomColsVar[i].split(":")[1] + "_" + nomCols[i + 1])  # Y
                n_vars.append(nomColsVar[i].split(":")[1] + "_" + nomCols[i + 2])  # Z
            elif "EMG" in nomCols[i]:
                # print(nomCols[i], nomColsVar[i])
                n_vars.append(nomColsVar[i].split(":")[1] + "_" + nomCols[i])
            # else:
            # print(nomCols[i])

    # [i for i in nomColsVar if "'" in i]
    # nomColsVar = [i for i in nomColsVar if "'" not in i]

    # carga todos los datos
    # CON GENFROMTXT FALLA SI NO EMPIEZA LA PRIMERA LÍNEA CON DATOS
    # provisional= np.genfromtxt(file_name, skip_header= ini_section+5, max_rows=end_section-ini_section-1, delimiter=sep, missing_values='', filling_values=np.nan, invalid_raise=True)
    # provisional=provisional[:, :len(n_vars)] #recorta solo hasta las variables

    # Convierte los datos en pandas dataframe. Pasa solo los que no son de velocidad o aceleración
    # dfReturn = pd.DataFrame(provisional[:, :len(n_vars)], columns=n_vars)
    # dfReturn = dfReturn.iloc[:, :len(n_vars)] #se queda solo con las columnas de las variables, quita las de velocidad si las hay

    # Con pandas directamente funciona (para evitar error si primera línea no son datos, lee la fila de las unidades y luego la quita)
    dfReturn = (
        pd.read_csv(
            file,
            delimiter=sep,
            header=None,
            skiprows=ini_section + 4,
            skipfooter=file_end - end_section - 5,
            usecols=range(len(n_vars)),
            engine="python",
        )
        .drop(index=0)
        .reset_index(drop=True)
        .astype(float)
    )

    # x=pd.read_csv(file, delimiter=sep, header=ini_section, skipfooter=file_end-end_section-5, engine='python')
    # x.columns
    # sub_nom_cols = x.iloc[0,:]

    # Nombra encabezado
    if section == "Devices":
        if "Noraxon Ultium" in n_vars[3]:
            var = [s.split("- ")[-1] for s in n_vars]
            coord = nomCols[: len(n_vars)]
            dfReturn.columns = var
            dimensions = ["n_var", "time"]
            var_name = ["n_var"]

    else:
        var = [
            "_".join(s.split("_")[:-1]) for s in n_vars[: len(n_vars)]
        ]  # gestiona si la variable tiene sep '_', lo mantiene
        coord = [
            s.split(":")[-1].lower() for s in nomCols[: len(n_vars)]
        ]  # pasa coordenadas a minúscula
        dfReturn.columns = pd.MultiIndex.from_tuples(
            list(zip(*[var, coord])), names=["n_var", "axis"]
        )
        # Elimina columnas con variables modeladas EMG si las hay
        dfReturn = dfReturn.drop(columns=dfReturn.filter(regex="emg"))
        # dfReturn.duplicated().sum()
        dimensions = ["n_var", "axis", "time"]
        var_name = ["n_var", "axis"]

    # dfReturn.columns=[var, coord]
    # dfReturn.columns.set_names(names=['Variable', 'Coord'], level=[0,1], inplace=True)

    # Incluye columna con tiempo
    dfReturn.insert(2, "time", np.arange(len(dfReturn)) / freq)

    if n_vars_load:
        dfReturn = dfReturn.loc[:, n_vars_load]

    if not raw:


        # if section == 'Devices':
        #    dfReturn.columns = dfReturn.columns.droplevel(1)

        # dfReturn.iloc[:,2:].melt(id_vars='time', var_name=['n_var', 'axis']).set_index(dimensions).to_xarray().to_array()

        # TODO: FIX THIS, DOES NOT WORK
        print("to_dataArray not implemented yet.\nTry with the Polars version")
        try:
            daReturn = (
                dfReturn.iloc[:, 2:]
                # .assign(**{'time':np.arange(len(dfReturn))/freq})
                .melt(id_vars="time", var_name=var_name)
                .set_index(dimensions)
                .to_xarray()
                .to_array()
                .squeeze("variable")
                .drop_vars("variable")  # la quita de dimensions y coordenadas
            )
            daReturn.name = section
            daReturn.attrs["freq"] = freq
            daReturn.time.attrs["units"] = "s"
        except:
            daReturn = xr.DataArray()

    if header_format == "flat" and section != "Devices":
        dfReturn.columns = dfReturn.columns.map("_".join).str.strip()

    # #Elimina las columnas de velocidad y aceleración, si las hay
    # borrarColsVA = dfReturn.filter(regex='|'.join(["'", "''"])).columns
    # dfReturn = dfReturn.drop(columns=borrarColsVA)

    # Si hace falta lo pasa a xArray
    if False:  # to_dataarray:
        # prueba para hacerlo directamente desde dataframe
        # dfReturn.assign(**{'time':np.arange(len(dfReturn))/freq}).drop(columns='').melt(id_vars='time').set_index(['n_var', 'axis', 'time']).to_xarray().to_array()

        if header_format != "flat":
            dfReturn.columns = dfReturn.columns.map("_".join).str.strip()

        # transforma los datos en xarray
        x = dfReturn.filter(regex="|".join(["_x", "_X"])).to_numpy().T
        y = dfReturn.filter(regex="|".join(["_y", "_Y"])).to_numpy().T
        z = dfReturn.filter(regex="|".join(["_z", "_Z"])).to_numpy().T
        data = np.stack([x, y, z])

        # Quita el identificador de la coordenada del final
        n_vars = dfReturn.filter(regex="|".join(["_x", "_X"])).columns.str.rstrip(
            "|".join(["_x", "_X"])
        )

        time = np.arange(x.shape[1]) / freq
        coords = {}
        coords["axis"] = ["x", "y", "z"]
        coords["n_var"] = n_vars
        coords["time"] = time

        daReturn = xr.DataArray(
            data=data,
            dims=("axis", "n_var", "time"),
            coords=coords,
            name=section,
            attrs={"freq": freq},
            # **kwargs,
        )
        if header_format != "flat":  # si hace falta lo vuelve a poner como multiindex
            dfReturn.columns = pd.MultiIndex.from_tuples(
                list(zip(*[var, coord])), names=["n_var", "axis"]
            )

    if to_dataarray and return_freq:
        return dfReturn, daReturn, freq
    elif to_dataarray:
        return dfReturn, daReturn
    elif return_freq:
        return dfReturn, freq
    else:
        return dfReturn


# =============================================================================
# %% TESTS
# =============================================================================
if __name__ == "__main__":

    # from biomdp.io.read_vicon_csv import read_vicon_csv
    work_path = Path(r"src\datasets")
    file = work_path / "vicon_CMJ_kinem_kinet_emg.csv"
    daData = read_vicon_csv(file, section="Trajectories")
    daData.plot.line(x="time", col="n_var", col_wrap=3)

    daData = read_vicon_csv(file, section="Model Outputs")
    daData.plot.line(x="time", col="n_var", col_wrap=3, sharey=False)

    daData = read_vicon_csv(file, section="Forces")
    daData.plot.line(x="time", col="n_var", col_wrap=3, sharey=False)

    daData = read_vicon_csv(file, section="EMG")
    daData.plot.line(x="time", col="n_var", col_wrap=3, sharey=False)

    import timeit

    def test_performance():
        result = read_vicon_csv(
            file, section="Model Outputs", engine="polars", raw=False
        )
        return result

    print(
        f'{timeit.timeit("test_performance()", setup="from __main__ import test_performance", number=10):.4f} s'
    )

    def test_performance():
        result = read_vicon_csv(file, section="Model Outputs", engine="pandas")
        return result

    print(
        f'{timeit.timeit("test_performance()", setup="from __main__ import test_performance", number=10):.4f} s'
    )



# -*- coding: utf-8 -*-
#!/usr/bin/env python

"""
Applies Butterworth filter. You can pass 1D array, pandas 2D DataFrame
or xarray dataarray.
Low or high pass function and bandpass function.
"""


# =============================================================================
# %% LOAD MODULES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "1.6.2"
__date__ = "17/02/2025"


"""
Updates:
    05/03/2025, v1.6.2
        - Adapted to biomdp with translations.
    
    17/02/2025, v1.6.1
        - Si se pasa un dataarray con atributo freq y sin parámetro fr,
          lo utiliza.
        - Incluidos tipos en parámetros funciones
    
    23/02/2024, v1.6.0
        - Introducida función plot con xarray.
    
    22/03/2023, v1.5.4
        - En filtro band pass corregido de lfilt a filtfilt.
        
    08/01/2023, v1.5.3
        - Ahora cuando se pasa un dataarray conserva el tipo de datos original.
    
    03/08/2022, v1.5.2
        - Ahora cuando se pasa un dataarray conserva los atributos.
        
    24/02/2022, v1.5.1
        - Cambiado el color del gráfico del original para que no se confundan
        
    08/05/2021, v1.5.0
        - Arreglado con xarray. Si tiene nan los rellena interpolando y después los elimina
        - Si no se pide el RMS o hacer la gráfica, no lo calcula.
        - Cambiados nombres de argumentos a más pythonics.
"""

from typing import Any
import numpy as np
import pandas as pd
import xarray as xr
import scipy.signal


# =============================================================================
# %% Functions
# =============================================================================
def filter_butter(
    dat_orig: np.ndarray | pd.DataFrame | xr.DataArray,
    fr: float | int | None = None,
    fc: float | int | None = None,
    order: float | int = 2.0,
    kind: str = "low",
    returnRMS: bool = False,
    show: bool = False,
    ax: Any | None = None,
) -> np.ndarray | pd.DataFrame | xr.DataArray:
    """
    Applies a Butterworth pass filter to data.

    Parameters
    ----------
    dat_orig : array_like, numpy, pandas or dataarray
        Original data to filter.
    fr : float, optional
        Sample frequency. Defaults to None.
    fc : float, optional
        Cut-off frequency. Defaults to None.
    order : float, optional
        Filter order. Defaults to 2.0.
    kind : str, optional
        Filter type. 'low' or 'high'. Defaults to 'low'.
    returnRMS : bool, optional
        Whether to return the root mean square (RMS) of the difference between
        the original and filtered data. Defaults to False.
    show : bool, optional
        Whether to display a plot of the original and filtered data. Defaults to False.
    ax : matplotlib axes, optional
        Axes on which to plot the data. Defaults to None.

    Returns
    -------
    filtData : array_like
        Filtered data.
    RMS : float, optional
        Root mean square of the difference between the original and filtered data.

    Notes
    -----
    Describe 2nd order, 2-pass filter as "double 2nd order Butterworth filter"
    (van den Bogert) http://biomch-l.isbweb.org/threads/26625-Matlab-Code-for-EMG-processing-(negative-deflection-after-normalization!!!)?p=32073#post32073


    Examples
    --------
    >>> import numpy as np
    >>> from biomdp.filter_butter import filter_butter
    >>> y = np.cumsum(np.random.randn(1000))
    >>> fy = filter_butter(y, fr=1000, fc=10, order=2, show=True)
    >>>
    >>> dfWalks = pd.DataFrame((np.random.random([100, 4])-0.5).cumsum(axis=0), columns=['A','B','C','D'])
    >>> dfWalks_Filt, RMS = filter_butter(dfWalks, 1000, 50, 2, show=True, returnRMS=True)

    """
    RMS = []

    # order = 2 #order 2 so that when doing the double pass it is 4th order
    passes = 2.0  # number of passes, forward and backward

    if fr is None:
        if isinstance(dat_orig, xr.DataArray) and "freq" in dat_orig.attrs:
            fr = dat_orig.freq
        else:
            if not dat_orig.isnull().all():
                fr = (
                    np.round(
                        1 / (dat_orig["time"][1] - dat_orig["time"][0]),
                        1,
                    )
                ).data
            else:
                raise RuntimeError("The sampling frequency (fr) must be specified.")

    # fc = 15
    Cf = (2 ** (1 / passes) - 1) ** (
        1 / (2 * order)
    )  # correction factor. Para 2nd order = 0.802 (Winter, 2009, p69)
    Wn = 2 * fc / fr / Cf

    b, a = scipy.signal.butter(order, Wn, btype=kind)

    # Pandas dataframe
    if isinstance(dat_orig, pd.DataFrame):
        dat_filt = pd.DataFrame()

        for i in range(dat_orig.shape[1]):
            dat_filt[dat_orig.columns[i]] = scipy.signal.filtfilt(
                b, a, dat_orig.iloc[:, i]
            )
        dat_filt.index = (
            dat_orig.index
        )  # esto es necesario por si se pasa un slice del dataframe

        if returnRMS or show == True:
            RMS = pd.DataFrame()
            for i in range(dat_orig.shape[1]):
                RMS.at[0, dat_orig.columns[i]] = np.linalg.norm(
                    dat_filt.iloc[:, i].values - dat_orig.iloc[:, i].values
                ) / np.sqrt(len(dat_orig.iloc[:, i]))

    # Pandas series
    elif isinstance(dat_orig, pd.Series):
        dat_filt = pd.Series(
            scipy.signal.filtfilt(b, a, dat_orig),
            index=dat_orig.index,
            name=dat_orig.name,
        )

        if returnRMS or show == True:
            RMS = np.linalg.norm(dat_filt - dat_orig) / np.sqrt(len(dat_orig))

    # Xarray dataarray
    elif isinstance(dat_orig, xr.DataArray):
        # dat_filt = xr.apply_ufunc(scipy.signal.filtfilt, b, a, dat_orig.dropna(dim='time')) #se asume que hay una dimensión tiempo
        dat_filt = xr.apply_ufunc(
            scipy.signal.filtfilt,
            b,
            a,
            dat_orig.interpolate_na(
                dim="time", method="linear", fill_value="extrapolate"
            ),
        )  # pad with nan the interpolated data
        # Retrieves the original data number by filling in with nan the endings as the original one
        dat_filt = dat_filt.where(xr.where(np.isnan(dat_orig), False, True), np.nan)
        dat_filt.attrs = dat_orig.attrs
        dat_filt = dat_filt.astype(dat_orig.dtype)

        if returnRMS or show == True:
            RMS = pd.DataFrame()
            for i in range(dat_orig.shape[0]):
                RMS.at[0, i] = np.linalg.norm(
                    dat_filt[i, :] - dat_orig[i, :]
                ) / np.sqrt(len(dat_orig[i, :]))
                # xr.apply_ufunc(np.linalg.norm, dat_filt[0,:], dat_orig[0,:])

    # Other data types
    else:
        dat_filt = scipy.signal.filtfilt(b, a, dat_orig)

        if returnRMS or show == True:
            RMS = np.linalg.norm(dat_filt - dat_orig) / np.sqrt(len(dat_orig))

    if show:
        _plot(dat_orig, dat_filt, RMS, fc, ax)

    if returnRMS:
        return dat_filt, RMS
    else:
        return dat_filt


# =============================================================================


# =============================================================================
# Shows plot
# =============================================================================
def _plot(
    dat_orig: np.ndarray | pd.DataFrame | xr.DataArray,
    dat_filt: np.ndarray | pd.DataFrame | xr.DataArray,
    RMS: float,
    fc: float,
    ax,
) -> None:
    import matplotlib.pyplot as plt

    if isinstance(dat_orig, xr.DataArray):

        def _xr_plot(orig, filt):
            plt.plot(filt, "b-", lw=0.8, label="Filt")
            plt.plot(orig, "r:", lw=1, label="Original")
            plt.legend()
            plt.show()

        _ = xr.apply_ufunc(
            _xr_plot,
            dat_orig,
            dat_filt,
            input_core_dims=[["time"], ["time"]],
            vectorize=True,
        )

    else:
        bNecesarioCerrarFigura = False

        if ax is None:
            bNecesarioCerrarFigura = True
            fig, ax = plt.subplots(1, 1, figsize=(10, 4))

        # Pandas dataframe
        if isinstance(dat_orig, pd.DataFrame):
            import seaborn as sns

            cmap = sns.color_palette("bright", n_colors=dat_orig.shape[1])
            dat_filt.plot(color=cmap, legend=False, ax=ax)
            dat_orig.plot(color=cmap, alpha=0.6, linestyle=":", legend=False, ax=ax)
            labels = [
                dat_orig.columns[x] + ", RMSE=" + "{:.3f}".format(RMS.iloc[0, x])
                for x in range(dat_orig.shape[1])
            ]
            plt.legend(labels)

        # Non pandas dataframe
        else:
            ax.plot(dat_filt, "b-", label="Filt (RMSE={:.3f})".format(RMS))
            ax.plot(dat_orig, "r:", alpha=0.8, label="Original")
            plt.legend(loc="best")

        ax.set_xlabel("Num. datos")
        ax.set_ylabel("Variable")
        ax.set_title("Filter Butterworth {0:3g} Hz".format(fc))

        if bNecesarioCerrarFigura:
            plt.show()


def filter_butter_bandpass(
    dat_orig: np.ndarray | pd.DataFrame | xr.DataArray,
    fr: float | int | None = None,
    fclow: float | int | None = None,
    fchigh: float | int | None = None,
    order: float | int = 2.0,
    show: bool = False,
    ax: Any | None = None,
) -> np.ndarray | pd.DataFrame | xr.DataArray:
    """
    Applies a Butterworth bandpass filter to data.

    Parameters
    ----------
    dat_orig : array_like, numpy, pandas or dataarray
        Original data to filter.
    fr : float, optional
        Sample frequency. Defaults to None.
    fclow, fchigh : float, optional
        Cut-off frequencies. Defaults to None.
    order : float, optional
        Filter order. Defaults to 2.0.
    show : bool, optional
        Whether to display a plot of the original and filtered data. Defaults to False.
    ax : matplotlib axes, optional
        Axes on which to plot the data. Defaults to None.

    Returns
    -------
    filtData : array_like
        Filtered data.

    Notes
    -----
    Información about bandpass Butterworth filter in:
    https://scipy-cookbook.readthedocs.io/items/ButterworthBandpass.html

    Examples
    --------
    >>> import numpy as np
    >>> from biomdp.filter_butter import filter_butter
    >>> y = np.cumsum(np.random.randn(1000))
    >>> fy = filter_butter(y, fr=1000, fc=10, order=2, show=True)
    >>>
    >>> dfWalks = pd.DataFrame((np.random.random([100, 4])-0.5).cumsum(axis=0), columns=['A','B','C','D'])
    >>> dfWalks_filt, RMS = filter_butter(dfWalks, 1000, 50, 2, show=True, returnRMS=True)

    """

    nyq = 0.5 * fr
    low = fclow / nyq
    high = fchigh / nyq
    b, a = scipy.signal.butter(order, [low, high], btype="band")

    # Pandas dataframe
    if isinstance(dat_orig, pd.DataFrame):
        dat_filt = pd.DataFrame()
        RMS = pd.DataFrame()
        for i in range(dat_orig.shape[1]):
            dat_filt[dat_orig.columns[i]] = scipy.signal.filtfilt(
                b, a, dat_orig.iloc[:, i]
            )
        dat_filt.index = (
            dat_orig.index
        )  # this is necessary in case a slice of the dataframe is passed

    # Pandas series
    elif isinstance(dat_orig, pd.Series):
        dat_filt = pd.Series(
            scipy.signal.filtfilt(b, a, dat_orig),
            index=dat_orig.index,
            name=dat_orig.name,
        )

    # Xarray dataarray
    elif isinstance(dat_orig, xr.DataArray):
        # dat_filt = xr.apply_ufunc(scipy.signal.filtfilt, b, a, dat_orig.dropna(dim='time')) #se asume que hay una dimensión tiempo
        dat_filt = xr.apply_ufunc(
            scipy.signal.filtfilt,
            b,
            a,
            dat_orig.interpolate_na(
                dim="time", method="linear", fill_value="extrapolate"
            ),
        )  # rellena los nan con datos interpolados
        dat_filt = dat_filt.where(
            xr.where(np.isnan(dat_orig), False, True), np.nan
        )  # recupera el nº de datos original rellenando con nan los finales como el original

    else:  # si los datos no son pandas dataframe
        dat_filt = scipy.signal.filtfilt(b, a, dat_orig)

    if show:
        _plot(dat_orig, dat_filt, RMS, fclow, ax)

    return dat_filt


# =============================================================================


# =============================================================================
# %% TESTS
# =============================================================================
if __name__ == "__main__":
    import matplotlib.pyplot as plt

    np.random.seed(2)
    y = np.cumsum(np.random.randn(1000))
    fy, rms = filter_butter(y, 1000, 10, 2, show=True, returnRMS=True)
    fy2, rms2 = filter_butter(y[100:300], 1000, 10, 2, show=True, returnRMS=True)

    fig, ax = plt.subplots(figsize=(8, 4))
    fig.suptitle("Big title", y=1.03)

    fy = filter_butter(y, 1000, 50, 2, show=True, ax=ax)
    ax.set_title("Little title", y=1.0)
    plt.show()

    # Multicolumn dataframe
    num = 1000
    colNames = ["A", "B", "C", "D"]
    dfWalks = pd.DataFrame(
        (np.random.random([num, 4]) - 0.5).cumsum(axis=0), columns=colNames
    )

    dfWalks_filt = filter_butter(dfWalks, 1000, 5, 2, show=True)
    dfWalks_filt, RMS = filter_butter(dfWalks, 1000, 50, 2, show=True, returnRMS=True)

    # Pandas series
    dfWalks_filt, RMS = filter_butter(
        dfWalks.iloc[:, 0], 1000, 5, 2, show=True, returnRMS=True
    )
    dfWalks_filt, RMS = filter_butter(
        dfWalks["A"], 1000, 50, 2, show=True, returnRMS=True
    )

    # %% Noisy wave
    t = np.arange(0, 2, 1 / 1000)
    # offset vertical
    of = [0, 0, 0, 0]
    # ampitudes
    a = [3, 0.5, 5, 0.3]
    # frecuencias
    f = [1, 60, 3, 40]
    # phase angle
    pa = [0, 0, 0, 0]
    waves = pd.DataFrame(
        np.array(
            [of[i] + a[i] * np.sin(2 * np.pi * f[i] * t + pa[i]) for i in range(len(a))]
        ).T
    )

    Wave = pd.DataFrame({"Wave1": waves[0] + waves[1], "Wave2": waves[2] + waves[3]})

    dfWave_filt = filter_butter(Wave, 1000, 10, 2, show=True)

    # With index change
    dfWave_filtCacho = filter_butter(Wave[100:300], 1000, 20, 2, show=True)

    dfWave_filtCacho, RMS = filter_butter(
        Wave.iloc[100:300, 0], 1000, 20, 2, show=True, returnRMS=True
    )

    fig, ax = plt.subplots(figsize=(8, 4))
    fy = filter_butter(Wave.iloc[400:600, 0], 1000, 50, 2, show=True, ax=ax)
    ax.set_title("(Sup little title)", y=1.0)
    plt.suptitle("Big title", y=1.03)
    plt.show()

    # %% Bandpass tests
    # Filter a noisy signal.
    fs = 5000.0
    lowcut = 500.0
    highcut = 1250.0

    T = 0.05
    nsamples = T * fs
    t = np.linspace(0, T, int(nsamples), endpoint=False)
    a = 0.02
    f0 = 600.0  # principal frequency to extract
    x = 0.1 * np.sin(2 * np.pi * 1.2 * np.sqrt(t))
    x += 0.01 * np.cos(2 * np.pi * 312 * t + 0.1)
    x += a * np.cos(2 * np.pi * f0 * t + 0.11)
    x += 0.03 * np.cos(2 * np.pi * 2000 * t)

    xFiltBand = filter_butter_bandpass(
        x, fs, lowcut, highcut, order=6, show=False, ax=None
    )

    fig, ax = plt.subplots(figsize=(8, 4))
    ax.plot(t, x, "b--")
    ax.plot(t, xFiltBand, "r")
    plt.hlines([-a, a], 0, T, "r", linestyles="--")
    plt.title("Bandpass Filter")
    plt.show()

    ###############################
    # %%xarray
    t = np.arange(0, 2, 1 / 1000)
    # vertical offset
    of = [0, 0, 0, 0]
    # ampitudes
    a = [3, 0.5, 5, 0.3]
    # frequencies
    f = [1, 60, 3, 40]
    # phase angle
    pa = [0, 0, 0, 0]
    waves = pd.DataFrame(
        np.array(
            [of[i] + a[i] * np.sin(2 * np.pi * f[i] * t + pa[i]) for i in range(len(a))]
        ).T
    )

    Wave = pd.DataFrame({"Wave1": waves[0] + waves[1], "Wave2": waves[2] + waves[3]})

    da = xr.DataArray(
        data=np.array(Wave).T,
        dims=["channel", "time"],
        coords={
            "channel": Wave.columns,
            "time": np.arange(0, len(Wave) / 1000, 1 / 1000),
        },
    )
    o = da.isel(channel=-1)
    da.plot.line(x="time")  # unfiltered
    da.isel(channel=1).plot()
    plt.show()

    np.linalg.norm(da.isel(channel=1) - da.isel(channel=0)) / np.sqrt(
        len(da.isel(channel=0))
    )

    oFilt, RMSEda = filter_butter(
        da, fr=1000, fc=10, order=2, returnRMS=True, show=False
    )
    da.plot.line(x="time")  # unfiltered
    oFilt.plot.line(x="time")  # filtered
    plt.show()

    # Compared with pandas is the same
    dfWave_filt, RMSEdf = filter_butter(
        Wave, fr=1000, fc=10, order=2, returnRMS=True, show=True
    )

    da.attrs["freq"] = 1000
    oFilt = filter_butter(da, fc=10, order=2, returnRMS=False, show=True)



# -*- coding: utf-8 -*-
"""
Created on Fry Mar 8 13:15:58 2024

@author: jose.lopeze

Reading of .csv files exported from Noraxon with IMU.
"""


# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "0.1.1"
__date__ = "13/03/2025"

"""
Updates:
    13/03/2025, v0.1.1
        - Adapted to biomdp with translations.

    08/03/2024, v0.1.0
        - Empezado tomando trozos sueltos.
            
"""

from typing import List
import numpy as np
import pandas as pd
import xarray as xr
import polars as pl

# import matplotlib.pyplot as plt

# from matplotlib.backends.backend_pdf import PdfPages #para guardar gráficas en pdf
# import seaborn as sns

from pathlib import Path
import time  # para cuantificar tiempos de procesado


# import sys
# sys.path.append('F:\Programacion\Python\Mios\Functions')
# #sys.path.append('G:\Mi unidad\Programacion\Python\Mios\Functions')

# from readViconCsv import read_vicon_csv


# =============================================================================
# %% Processing functions
# =============================================================================


def split_dim_axis(da: xr.DataArray) -> xr.DataArray:
    # Separa el xarray en ejes creando dimensión axis
    if "Accel " in da.n_var.to_series().iloc[0]:
        sensor_type = "Accel "
    elif "Gyro " in da.n_var.to_series().iloc[0]:
        sensor_type = "Gyro "
    elif "Mag " in da.n_var.to_series().iloc[0]:
        sensor_type = "Mag "
    else:
        raise Exception('Sensor type not found, try "Accel", "Gyro" or "Mag"')

    x = da.sel(
        n_var=da.n_var.str.endswith(f"{sensor_type[0]}x")
    )  # .rename({"n_var": "axis"})
    x = x.assign_coords(
        n_var=[
            s[:-3] for s in x.n_var.to_series()
        ]  # [s.split(sensor_type)[1][:-3] for s in x.n_var.to_series()]
    )

    y = da.sel(
        n_var=da.n_var.str.endswith(f"{sensor_type[0]}y")
    )  # .rename({"n_var": "axis"})
    y = y.assign_coords(n_var=[s[:-3] for s in y.n_var.to_series()])

    z = da.sel(
        n_var=da.n_var.str.endswith(f"{sensor_type[0]}z")
    )  # .rename({"n_var": "axis"})
    z = z.assign_coords(n_var=[s[:-3] for s in z.n_var.to_series()])

    da = (
        xr.concat([x, y, z], dim="axis")
        # .assign_coords(n_var="plat1")
        .assign_coords(axis=["x", "y", "z"])
        # .expand_dims({"n_var": 1})
    )

    return da


def split_dim_side(da: xr.DataArray) -> xr.DataArray:
    # Separa el xarray en ejes creando dimensión axis

    L = da.sel(n_var=da.n_var.str.contains("izquierda"))
    L = L.assign_coords(n_var=[s.split(" izquierda")[0] for s in L.n_var.to_series()])

    R = da.sel(n_var=da.n_var.str.contains("derecha"))  # .rename({"n_var": "axis"})
    R = R.assign_coords(n_var=[s.split(" derecha")[0] for s in R.n_var.to_series()])

    da = (
        xr.concat([L, R], dim="side")
        # .assign_coords(n_var="plat1")
        .assign_coords(side=["L", "R"])
        # .expand_dims({"n_var": 1})
    )

    return da


def assign_subcategories_xr(
    da: xr.DataArray, n_project: str | None = None
) -> xr.DataArray:

    if len(da.ID.to_series().iloc[0].split("_")) == 5:
        da = da.assign_coords(
            estudio=(
                n_project
            ),  # "ID", da.ID.to_series().str.split("_").str[0].to_list()),
            particip=("ID", da.ID.to_series().str.split("_").str[1].to_list()),
            tipo=("ID", da.ID.to_series().str.split("_").str[2].to_list()),
            subtipo=("ID", da.ID.to_series().str.split("_").str[3].to_list()),
            repe=("ID", da.ID.to_series().str.split("_").str[4].to_list()),
        )

    elif len(da.ID.to_series().iloc[0].split("_")) == 4:
        if n_project is None:
            n_project = "X"
        da = da.assign_coords(
            estudio=("ID", [n_project] * len(da.ID)),
            particip=("ID", da.ID.to_series().str.split("_").str[0].to_list()),
            tipo=("ID", da.ID.to_series().str.split("_").str[1].to_list()),
            subtipo=("ID", da.ID.to_series().str.split("_").str[2].to_list()),
            repe=("ID", da.ID.to_series().str.split("_").str[3].to_list()),
        )

    """
    #versión basada en df polars
    da = da.assign_coords(estudio=('ID', df.filter(pl.col('time')==0.000).get_column('estudio').to_list()),
                                              particip=('ID', df.filter(pl.col('time')==0.000).get_column('particip').to_list()),
                                              tipo=('ID', df.filter(pl.col('time')==0.000).get_column('tipo').to_list()),
                                              subtipo=('ID', df.filter(pl.col('time')==0.000).get_column('subtipo').to_list()),
                                              repe=('ID', df.filter(pl.col('time')==0.000).get_column('repe').to_list()),
                                              )
    """
    """
    #Versión antigua si no se sabe si hay dimensión repe o no.
    #En la versión actual no hay dimensión repe, se guarda en el ID
    est=[] #un poco complejo, pero si no todos tienen repe=1 no funcionaba bien
    tip=[]
    subtip=[]
    partic=[]
    repe=[]
    #da.ID.str.split(dim='splt', sep='_')
    for n in da.ID:
        #if len(n.str.split('_'))
        partes = n.str.split(dim='splt', sep='_')
        if len(partes)==3:
            est.append(n_project)
            partic.append(partes.data[0])
            tip.append(partes.data[1])
            subtip.append(subtipo)
            repe.append(partes.data[-1])
        elif len(partes)==4:
            est.append(n_project)
            partic.append(partes.data[0])
            tip.append(partes.data[1])
            subtip.append(partes.data[2])
            repe.append(partes.data[-1])
        elif len(partes)==5:
            est.append(partes.data[0])
            partic.append(partes.data[1])
            tip.append(partes.data[2])
            subtip.append(partes.data[3])
            repe.append(partes.data[-1])
    
    da = da.assign_coords(estudio=('ID', est), particip=('ID', partic), tipo=('ID', tip), subtipo=('ID', subtip), repe=('ID', repe))
    """
    # if 'repe' in da.dims: #solo lo añade si no tiene ya la dimensión repe
    #     da = da.assign_coords(estudio=('ID', est), particip=('ID', partic), tipo=('ID', tip), subtipo=('ID', subtip))
    # else:
    #     da = da.assign_coords(estudio=('ID', est), particip=('ID', partic), tipo=('ID', tip), subtipo=('ID', subtip), repe=('ID', repe))

    return da


def df_to_da(
    dfAll: pl.DataFrame | pd.DataFrame, n_project: str | None = None, show: bool = False
) -> xr.DataArray:
    if isinstance(dfAll, pl.DataFrame):
        # Transforma df polars a dataarray con todas las variables cargadas
        vars_leidas = dfAll.select(
            pl.exclude(
                ["time", "estudio", "tipo", "subtipo", "ID", "particip", "repe"]
            ),
        ).columns

        dfpd = dfAll.melt(
            id_vars=["ID", "time"], value_vars=vars_leidas, variable_name="n_var"
        ).to_pandas()

    else:  # from pandas
        vars_leidas = dfAll.drop(
            columns=["time", "estudio", "tipo", "subtipo", "ID", "particip", "repe"]
        ).columns
        dfpd = dfAll.drop(
            columns=["estudio", "tipo", "subtipo", "particip", "repe"]
        ).melt(id_vars=["ID", "time"], var_name="n_var")

    daAll = (
        # dfpd.drop(columns=["estudio", "tipo", "subtipo", "particip", "repe"])
        dfpd  # .melt(id_vars=["ID", "time"], var_name="n_var")
        # pd.melt(dfAllArchivos.to_pandas().drop(columns=['estudio','tipo','subtipo']), id_vars=['ID', 'repe', 'time'], var_name='axis')
        .set_index(["ID", "n_var", "time"])
        .to_xarray()
        .to_array()
        .squeeze("variable")
        .drop_vars("variable")
    )

    # Assign extra coordinates
    daAll = assign_subcategories_xr(da=daAll, n_project=n_project)
    # daAll = daAll.assign_coords(estudio=('ID', dfAll.filter(pl.col('time')==0.000).get_column('estudio').to_list()),
    #                                          particip=('ID', dfAll.filter(pl.col('time')==0.000).get_column('particip').to_list()),
    #                                          tipo=('ID', dfAll.filter(pl.col('time')==0.000).get_column('tipo').to_list()),
    #                                          subtipo=('ID', dfAll.filter(pl.col('time')==0.000).get_column('subtipo').to_list()),
    #                                          repe=('ID', dfAll.filter(pl.col('time')==0.000).get_column('repe').to_list()),
    #                                          )
    # Ajusta tipo de coordenada tiempo, necesario??
    ###########daAllArchivos = daAllArchivos.assign_coords(time=('time', daAllArchivos.time.astype('float32').values))

    # daAllArchivos.sel(ID='PCF_SCT05', axis='z').plot.line(x='time', col='repe')
    # daAllArchivos.assign_coords(time=daAllArchivos.time.astype('float32'))

    # daAll.attrs = {
    #     "freq": (np.round(1 / (daAll.time[1] - daAll.time[0]), 1)).data,
    #     "units": "N",
    # }
    daAll.attrs["freq"] = (np.round(1 / (daAll.time[1] - daAll.time[0]), 1)).data
    daAll.time.attrs["units"] = "s"

    return daAll


def read_noraxon_pd(
    file: str | Path, n_vars_load: List[str] | None = None, to_dataarray: bool = False
) -> xr.DataArray | pl.DataFrame:
    # file = Path(r"F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\PilotoNoraxon\S000\2024-03-08-10-43_PO_S000_carrera_001.csv")
    df = pd.read_csv(
        file, skiprows=3, header=0, engine="c"
    ).drop(  # .astype(np.float64)
        columns=["Activity", "Marker"]
    )

    # df = df.drop(columns=df.filter(regex="(Normal)") + df.filter(regex="(Tiempo.)"))

    #       df.filter(regex="(Normal)").values)).drop(columns=df.filter(regex="(Tiempo.)"))
    # df.filter(regex="(Normal)")+df.filter(regex="(Tiempo.)")
    # df.dropna(axis="columns", how="all")

    # df.filter(regex="Tiempo.")

    if to_dataarray:
        print("_to_dataarray not implemented yet. Try using read_noraxon_pl")
        da = xr.DataArray()
        return da

    return df


def read_noraxon_pl(
    file: str | Path, n_vars_load: List[str] | None = None, to_dataarray: bool = False
) -> xr.DataArray | pl.DataFrame:
    # file = Path(r"F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\PilotoNoraxon\S000\2024-03-08-10-43_PO_S000_carrera_001.csv")

    df = (
        pl.read_csv(
            file,
            has_header=True,
            skip_rows=3,
            # skip_rows_after_header=0,
            columns=n_vars_load,
            # separator=",",
        )
        # .select(pl.exclude("^.*_duplicated_.*$"))  # quita columnas de tiempo duplicadas
        .select(pl.exclude(pl.String))  # quita columnas de texto con datos (Normal)
        # .with_columns(pl.all().cast(pl.Float64()))
    )

    """
    cadera = df.select(
            pl.col("^*Flexo-extensión cadera .*::Y$")
        )  # .to_numpy()
    """
    # df = df.to_pandas()

    # ----Transform polars to xarray
    if to_dataarray:

        # # Separa ejes articulares
        # x = df.select(pl.col("^Flexo-extensión.*$")).to_numpy()
        # y = df.select(pl.col("^Aducción-abducción.*$")).to_numpy()
        # z = df.select(pl.col("^Rotación.*$")).to_numpy()
        # data = np.stack([x, y, z])

        freq = 1 / (df[1, "time"] - df[0, "time"])

        # coords = {
        #     "axis": ["x", "y", "z"],
        #     "time": np.arange(data.shape[1]) / freq,
        #     "n_var": ["Force"],  # [x[:ending] for x in df.columns if 'x' in x[-1]],
        # }
        coords = {
            "time": np.arange(df.shape[0]) / freq,
            "n_var": df.columns,  # [x[:ending] for x in df.columns if 'x' in x[-1]],
        }
        da = (
            xr.DataArray(
                data=df.to_numpy(),
                dims=coords.keys(),
                coords=coords,
            )
            # .astype(float)
            # .transpose("n_var", "time")
        )
        da.name = "EMG"
        da.attrs["freq"] = freq
        da.time.attrs["units"] = "s"
        da.attrs["units"] = "mV"

        return da

    return df


def load_merge_noraxon_csv(
    path: str | Path,
    section: str = "EMG",
    n_project: str | None = None,
    data_type: str | None = None,
    show: bool = False,
) -> xr.DataArray:
    """
    Parameters
    ----------
    path : TYPE
        DESCRIPTION.
    n_vars_load : TYPE, optional
        DESCRIPTION. The default is None.
    n_project : string, optional
        DESCRIPTION. The name of the study.
    data_type:
        Conversión al tipo de datos indicado. Por defecto es None, que quiere
        decir que se mantiene el tipo original ('float64')
    show : bool, optional
        DESCRIPTION. The default is False.


    Returns
    -------
    daAllArchivos : xarray DataArray
        DESCRIPTION. DataArray with the loaded files concatenated.


    """
    if data_type is None:
        data_type = float

    files_list = sorted(
        list(path.glob("*.csv"))  # "**/*.csv"
    )  #'**/*.txt' incluye los que haya en subcarpetas
    files_list = [
        x
        for x in files_list
        if "error" not in x.name and "info" not in x.name and "_error" not in x.name
    ]  # selecciona archivos

    all_vars = [
        "Ultium EMG.EMG 1 (uV)",
        "Ultium EMG.Internal Accel 1 Ax (mG)",
        "Ultium EMG.Internal Accel 1 Ay (mG)",
        "Ultium EMG.Internal Accel 1 Az (mG)",
        "Ultium EMG.Internal Gyro 1 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 1 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 1 Gz (deg/s)",
        "Ultium EMG.Internal Mag 1 Mx (mGauss)",
        "Ultium EMG.Internal Mag 1 My (mGauss)",
        "Ultium EMG.Internal Mag 1 Mz (mGauss)",
        "Ultium EMG.EMG 2 (uV)",
        "Ultium EMG.Internal Accel 2 Ax (mG)",
        "Ultium EMG.Internal Accel 2 Ay (mG)",
        "Ultium EMG.Internal Accel 2 Az (mG)",
        "Ultium EMG.Internal Gyro 2 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 2 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 2 Gz (deg/s)",
        "Ultium EMG.Internal Mag 2 Mx (mGauss)",
        "Ultium EMG.Internal Mag 2 My (mGauss)",
        "Ultium EMG.Internal Mag 2 Mz (mGauss)",
        "Ultium EMG.EMG 3 (uV)",
        "Ultium EMG.Internal Accel 3 Ax (mG)",
        "Ultium EMG.Internal Accel 3 Ay (mG)",
        "Ultium EMG.Internal Accel 3 Az (mG)",
        "Ultium EMG.Internal Gyro 3 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 3 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 3 Gz (deg/s)",
        "Ultium EMG.Internal Mag 3 Mx (mGauss)",
        "Ultium EMG.Internal Mag 3 My (mGauss)",
        "Ultium EMG.Internal Mag 3 Mz (mGauss)",
        "Ultium EMG.EMG 4 (uV)",
        "Ultium EMG.Internal Accel 4 Ax (mG)",
        "Ultium EMG.Internal Accel 4 Ay (mG)",
        "Ultium EMG.Internal Accel 4 Az (mG)",
        "Ultium EMG.Internal Gyro 4 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 4 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 4 Gz (deg/s)",
        "Ultium EMG.Internal Mag 4 Mx (mGauss)",
        "Ultium EMG.Internal Mag 4 My (mGauss)",
        "Ultium EMG.Internal Mag 4 Mz (mGauss)",
        "Ultium EMG.EMG 5 (uV)",
        "Ultium EMG.Internal Accel 5 Ax (mG)",
        "Ultium EMG.Internal Accel 5 Ay (mG)",
        "Ultium EMG.Internal Accel 5 Az (mG)",
        "Ultium EMG.Internal Gyro 5 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 5 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 5 Gz (deg/s)",
        "Ultium EMG.Internal Mag 5 Mx (mGauss)",
        "Ultium EMG.Internal Mag 5 My (mGauss)",
        "Ultium EMG.Internal Mag 5 Mz (mGauss)",
        "Ultium EMG.EMG 6 (uV)",
        "Ultium EMG.Internal Accel 6 Ax (mG)",
        "Ultium EMG.Internal Accel 6 Ay (mG)",
        "Ultium EMG.Internal Accel 6 Az (mG)",
        "Ultium EMG.Internal Gyro 6 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 6 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 6 Gz (deg/s)",
        "Ultium EMG.Internal Mag 6 Mx (mGauss)",
        "Ultium EMG.Internal Mag 6 My (mGauss)",
        "Ultium EMG.Internal Mag 6 Mz (mGauss)",
        "Ultium EMG.Internal Accel 7 Ax (mG)",
        "Ultium EMG.Internal Accel 7 Ay (mG)",
        "Ultium EMG.Internal Accel 7 Az (mG)",
        "Ultium EMG.Internal Gyro 7 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 7 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 7 Gz (deg/s)",
        "Ultium EMG.Internal Mag 7 Mx (mGauss)",
        "Ultium EMG.Internal Mag 7 My (mGauss)",
        "Ultium EMG.Internal Mag 7 Mz (mGauss)",
        "Ultium EMG.Internal Accel 8 Ax (mG)",
        "Ultium EMG.Internal Accel 8 Ay (mG)",
        "Ultium EMG.Internal Accel 8 Az (mG)",
        "Ultium EMG.Internal Gyro 8 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 8 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 8 Gz (deg/s)",
        "Ultium EMG.Internal Mag 8 Mx (mGauss)",
        "Ultium EMG.Internal Mag 8 My (mGauss)",
        "Ultium EMG.Internal Mag 8 Mz (mGauss)",
        "Ultium EMG.Internal Accel 9 Ax (mG)",
        "Ultium EMG.Internal Accel 9 Ay (mG)",
        "Ultium EMG.Internal Accel 9 Az (mG)",
        "Ultium EMG.Internal Gyro 9 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 9 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 9 Gz (deg/s)",
        "Ultium EMG.Internal Mag 9 Mx (mGauss)",
        "Ultium EMG.Internal Mag 9 My (mGauss)",
        "Ultium EMG.Internal Mag 9 Mz (mGauss)",
        "Ultium EMG.Internal Accel 10 Ax (mG)",
        "Ultium EMG.Internal Accel 10 Ay (mG)",
        "Ultium EMG.Internal Accel 10 Az (mG)",
        "Ultium EMG.Internal Gyro 10 Gx (deg/s)",
        "Ultium EMG.Internal Gyro 10 Gy (deg/s)",
        "Ultium EMG.Internal Gyro 10 Gz (deg/s)",
        "Ultium EMG.Internal Mag 10 Mx (mGauss)",
        "Ultium EMG.Internal Mag 10 My (mGauss)",
        "Ultium EMG.Internal Mag 10 Mz (mGauss)",
    ]

    if section == "EMG":
        sect = "EMG "
    else:
        sect = section

    n_vars_load = ["time"] + [s for s in all_vars if sect in s]

    # file = Path("F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\PilotoNoraxon\S000\2024-03-08-10-43_PO_S000_carrera_001.csv")

    print("\nLoading files...")
    timerCarga = time.perf_counter()  # inicia el contador de tiempo

    num_processed_files = 0
    dfAll = (
        []
    )  # guarda los dataframes que va leyendo en formato lista y al final los concatena
    daAll = []
    error_files = []  # guarda los nombres de archivo que no se pueden abrir y su error
    for nf, file in enumerate(files_list):
        print(f"Cargando archivo nº {nf+1}/{len(files_list)}: {file.name}")
        try:
            timerSub = time.perf_counter()  # inicia el contador de tiempo

            dfProvis = read_noraxon_pl(file, n_vars_load)
            n_file = file.stem.split("_")[
                1:
            ]  # para quitar lo primero que pone, que es la fecha
            if len(n_file) == 5:
                n_project = n_file[0] if n_project is None else n_project
                particip = n_file[-4]
                tipo = n_file[-3]
                subtipo = n_file[-2]
            elif len(n_file) == 4:
                n_project = n_file[0]
                particip = n_file[1]
                tipo = n_file[2]
                subtipo = "X"
            elif len(n_file) == 3:
                particip = n_file[0]
                tipo = n_file[-2]
                subtipo = "X"
            if n_project is None:
                n_project = "EstudioX"

            repe = str(int(n_file[-1]))  # int(file.stem.split('.')[0][-1]
            ID = f"{particip}_{tipo}_{subtipo}_{repe}"  # f'{n_project}_{file.stem.split("_")[0]}_{tipo}_{subtipo}'

            # freq = np.round(1/(dfProvis['time'][1]-dfProvis['time'][0]),1)

            # Añade categorías
            dfProvis = dfProvis.with_columns(
                [
                    pl.lit(n_project).alias("estudio"),
                    pl.lit(tipo).alias("tipo"),
                    pl.lit(subtipo).alias("subtipo"),
                    pl.lit(ID).alias("ID"),
                    pl.lit(particip).alias("particip"),
                    pl.lit(repe).alias("repe"),
                ]
            )  # .select(['estudio', 'tipo', 'subtipo', 'ID', 'repe'] + nom_vars_cargar)

            dfAll.append(dfProvis)

            print(f"{dfAll[-1].shape[0]} rows and {dfAll[-1].shape[1]} columns")
            print("Time {0:.3f} s \n".format(time.perf_counter() - timerSub))
            num_processed_files += 1

        except Exception as err:  # Si falla anota un error y continúa
            print(
                f"\nATTENTION. Unable to process {file.parent.name}, {file.name}, {err}\n"
            )
            error_files.append(file.parent.name + " " + file.name + " " + str(err))
            continue

    dfAll = pl.concat(dfAll)

    print(
        f"Loaded {num_processed_files} files in {time.perf_counter()-timerCarga:.3f} s \n"
    )

    # Show file errors
    if len(error_files) > 0:
        print("\nATTENTION. Unable to load:")
        for x in range(len(error_files)):
            print(error_files[x])

    if isinstance(data_type, str):
        cast = pl.Float32() if data_type == "float32" else pl.Float64()
        dfAll = dfAll.select(
            # pl.col(['n_project', 'tipo', 'subtipo', 'ID', 'repe']),
            pl.exclude(n_vars_load),
            pl.col(n_vars_load).cast(cast),
        )

    # Rename files
    n_vars_load2 = [s.split(".")[-1] for s in n_vars_load[1:] if sect in s]

    # Rename columnas
    # if section == 'EMG':
    #     rename = [s.split('Internal ')[-1][:-5] for s in n_vars_load2]
    # elif section == 'Accel':
    #     rename = [s.split('Internal ')[-1][:-5] for s in n_vars_load2]
    # elif section == 'Gyro':
    rename = [s.split("Internal ")[-1].split(" (")[0] for s in n_vars_load2]

    dfAll = dfAll.rename(dict(zip(dfAll.columns[1 : len(rename) + 1], rename)))

    daAll = df_to_da(dfAll, n_project=n_project)
    daAll.name = section
    daAll.attrs["units"] = n_vars_load2[0].split("(")[-1].split(")")[0]

    # daAll = split_dim_axis(daAll)

    # daAll = split_dim_side(daAll)

    return daAll


# =============================================================================
# %% Prueba las funciones
# =============================================================================
if __name__ == "__main__":
    ###################################
    # PROBAR CÁLCULOS ÁNGULOS IMUS CON PYKINEMATICS
    # pip install pykinematics
    # TAMBIÉN PROBAR CON pip install pyjamalib, conda install conda-forge::pyjamalib
    # TAMBIÉN SCIKIT-KINEMATICS http://work.thaslwanter.at/skinematics/html/
    # pip install scikit-kinematics
    ###################################
    # work_path = Path('F:\Programacion\Python\Mios\TratamientoDatos\EjemploBikefitting')
    work_path = Path(
        r"F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\PilotoNoraxon\S000"
    )

    # Carga EMG
    daEMG = load_merge_noraxon_csv(path=work_path, section="EMG")
    # daEMG.attrs['units'] = 'uV'
    daEMG.plot.line(x="time", col="n_var", row="ID", sharey=False)

    # Load Acelerometer
    daAcc = load_merge_noraxon_csv(path=work_path, section="Accel")
    daAcc = split_dim_axis(da=daAcc)
    daAcc.plot.line(x="time", col="n_var", row="ID", sharey=False)

    # Load Gyroscope
    daGyro = load_merge_noraxon_csv(path=work_path, section="Gyro")
    daGyro = split_dim_axis(da=daGyro)
    daGyro.plot.line(x="time", col="n_var", row="ID", sharey=False)
    daGyro.where(daGyro.tipo == "salto", drop=True).isel(ID=0).plot.line(
        x="time", col="n_var", col_wrap=4, sharey=False
    )
    daGyro.sel(ID=daGyro.ID.str.contains("salto")).isel(ID=0).plot.line(
        x="time", col="n_var", col_wrap=4, sharey=False
    )

    # Load Magnetometer
    daMag = load_merge_noraxon_csv(path=work_path, section="Mag")
    daMag = split_dim_axis(da=daMag)
    daMag.plot.line(x="time", col="n_var", row="ID", sharey=False)
    daMag.sel(ID=daMag.ID.str.contains("salto")).isel(ID=0).plot.line(
        x="time", col="n_var", col_wrap=4, sharey=False
    )

    """
    file = Path(r"F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\PilotoNoraxon\S000\2024-03-08-10-43_PO_S000_carrera_001.csv")
    #Sigue siendo más rápido con polars que pd engine='c'
    t = time.perf_counter()  # inicia el contador de tiempo
    for i in range(10):
        dfProvis = read_noraxon_pd(file)
    print(time.perf_counter()-t)
    
    
    t = time.perf_counter()  # inicia el contador de tiempo
    for i in range(10):
        dfProvis = read_noraxon_pl(file, n_vars_load=['Ultium EMG.Internal Accel 1 Ax (mG)'], to_dataarray=False)
    print(time.perf_counter()-t)
    
    read_noraxon_pl(file, to_dataarray=False)
    """



# -*- coding: utf-8 -*-
"""
Created on Fry Mar 1 13:15:58 2024

@author: jose.lopeze

Lectura de archivos .csv exportados de iSen.

"""


#
# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "0.1.2"
__date__ = "28/03/2025"

"""
Updates:
    28/03/2025, v0.1.2
        - Indluded Polars import in df_to_da function.
        
    11/03/2025, v0.1.1
        - Adapted to biomdp with translations.
    
    17/08/2024, v0.1.0
        - Basado en el usado para registros 2021 para TFM de María Aracil. 

        """
from typing import List, Any
import numpy as np

import pandas as pd
import xarray as xr

# import polars as pl

import matplotlib.pyplot as plt

# from matplotlib.backends.backend_pdf import PdfPages #para guardar gráficas en pdf
# import seaborn as sns

from pathlib import Path
import time  # para cuantificar tiempos de procesado


# import sys
# sys.path.append('F:\Programacion\Python\Mios\Functions')
# #sys.path.append('G:\Mi unidad\Programacion\Python\Mios\Functions')

# from readViconCsv import read_vicon_csv


# =============================================================================
# %% Functions
# =============================================================================


def split_dim_axis(da: xr.DataArray) -> xr.DataArray:
    # Separa el xarray en ejes creando dimensión axis

    x = da.sel(
        n_var=da.n_var.str.contains("Flexo-extensión")
    )  # .rename({"n_var": "axis"})
    x = x.assign_coords(
        n_var=[s.split("Flexo-extensión ")[1][:-3] for s in x.n_var.to_series()]
    )

    y = da.sel(
        n_var=da.n_var.str.contains("Abducción-aducción")
    )  # .rename({"n_var": "axis"})
    y = y.assign_coords(
        n_var=[s.split("Abducción-aducción ")[1][:-3] for s in y.n_var.to_series()]
    )

    z = da.sel(n_var=da.n_var.str.contains("Rotación"))  # .rename({"n_var": "axis"})
    z = z.assign_coords(
        n_var=[s.split("Rotación ")[1][:-3] for s in z.n_var.to_series()]
    )

    da = (
        xr.concat([x, y, z], dim="axis")
        # .assign_coords(n_var="plat1")
        .assign_coords(axis=["x", "y", "z"])
        # .expand_dims({"n_var": 1})
    )

    return da


def split_dim_side(da: xr.DataArray) -> xr.DataArray:
    # Separa el xarray en ejes creando dimensión axis

    L = da.sel(n_var=da.n_var.str.contains("izquierda"))
    L = L.assign_coords(n_var=[s.split(" izquierda")[0] for s in L.n_var.to_series()])

    R = da.sel(n_var=da.n_var.str.contains("derecha"))  # .rename({"n_var": "axis"})
    R = R.assign_coords(n_var=[s.split(" derecha")[0] for s in R.n_var.to_series()])

    da = (
        xr.concat([L, R], dim="side")
        # .assign_coords(n_var="plat1")
        .assign_coords(side=["L", "R"])
        # .expand_dims({"n_var": 1})
    )

    return da


def assign_subcategories_xr(da: xr.DataArray, n_project: str = None) -> xr.DataArray:

    if len(da.ID.to_series().iloc[0].split("_")) == 5:
        da = da.assign_coords(
            estudio=(
                n_project
            ),  # "ID", da.ID.to_series().str.split("_").str[0].to_list()),
            particip=("ID", da.ID.to_series().str.split("_").str[1].to_list()),
            tipo=("ID", da.ID.to_series().str.split("_").str[2].to_list()),
            subtipo=("ID", da.ID.to_series().str.split("_").str[3].to_list()),
            repe=("ID", da.ID.to_series().str.split("_").str[4].to_list()),
        )

    elif len(da.ID.to_series().iloc[0].split("_")) == 4:
        if n_project is None:
            n_project = "X"
        da = da.assign_coords(
            estudio=("ID", [n_project] * len(da.ID)),
            particip=("ID", da.ID.to_series().str.split("_").str[0].to_list()),
            tipo=("ID", da.ID.to_series().str.split("_").str[1].to_list()),
            subtipo=("ID", da.ID.to_series().str.split("_").str[2].to_list()),
            repe=("ID", da.ID.to_series().str.split("_").str[3].to_list()),
        )

    """
    #versión basada en df polars
    da = da.assign_coords(estudio=('ID', df.filter(pl.col('time')==0.000).get_column('estudio').to_list()),
                                              particip=('ID', df.filter(pl.col('time')==0.000).get_column('particip').to_list()),
                                              tipo=('ID', df.filter(pl.col('time')==0.000).get_column('tipo').to_list()),
                                              subtipo=('ID', df.filter(pl.col('time')==0.000).get_column('subtipo').to_list()),
                                              repe=('ID', df.filter(pl.col('time')==0.000).get_column('repe').to_list()),
                                              )
    """
    """
    #Versión antigua si no se sabe si hay dimensión repe o no.
    #En la versión actual no hay dimensión repe, se guarda en el ID
    est=[] #un poco complejo, pero si no todos tienen repe=1 no funcionaba bien
    tip=[]
    subtip=[]
    partic=[]
    repe=[]
    #da.ID.str.split(dim='splt', sep='_')
    for n in da.ID:
        #if len(n.str.split('_'))
        partes = n.str.split(dim='splt', sep='_')
        if len(partes)==3:
            est.append(n_project)
            partic.append(partes.data[0])
            tip.append(partes.data[1])
            subtip.append(subtipo)
            repe.append(partes.data[-1])
        elif len(partes)==4:
            est.append(n_project)
            partic.append(partes.data[0])
            tip.append(partes.data[1])
            subtip.append(partes.data[2])
            repe.append(partes.data[-1])
        elif len(partes)==5:
            est.append(partes.data[0])
            partic.append(partes.data[1])
            tip.append(partes.data[2])
            subtip.append(partes.data[3])
            repe.append(partes.data[-1])
    
    da = da.assign_coords(estudio=('ID', est), particip=('ID', partic), tipo=('ID', tip), subtipo=('ID', subtip), repe=('ID', repe))
    """
    # if 'repe' in da.dims: #solo lo añade si no tiene ya la dimensión repe
    #     da = da.assign_coords(estudio=('ID', est), particip=('ID', partic), tipo=('ID', tip), subtipo=('ID', subtip))
    # else:
    #     da = da.assign_coords(estudio=('ID', est), particip=('ID', partic), tipo=('ID', tip), subtipo=('ID', subtip), repe=('ID', repe))

    return da


def df_to_da(
    dfAll: pd.DataFrame, n_project: str | None = None, show: bool = False
) -> xr.DataArray:
    try:
        import polars as pl
    except:
        raise ImportError(
            "Polars package not instaled. Install it if you want to use the accelerated version"
        )

    if isinstance(dfAll, pl.DataFrame):
        # Transforma df polars a dataarray con todas las variables cargadas
        vars_leidas = dfAll.select(
            pl.exclude(
                ["time", "estudio", "tipo", "subtipo", "ID", "particip", "repe"]
            ),
        ).columns

        dfpd = dfAll.unpivot(
            index=["ID", "time"], on=vars_leidas, variable_name="n_var"
        ).to_pandas()

    else:  # viene con pandas
        vars_leidas = dfAll.drop(
            columns=["time", "estudio", "tipo", "subtipo", "ID", "particip", "repe"]
        ).columns
        dfpd = dfAll.drop(
            columns=["estudio", "tipo", "subtipo", "particip", "repe"]
        ).melt(id_vars=["ID", "time"], var_name="n_var")
    daAll = (
        # dfpd.drop(columns=["estudio", "tipo", "subtipo", "particip", "repe"])
        dfpd  # .melt(id_vars=["ID", "time"], var_name="n_var")
        # pd.melt(dfAllFiles.to_pandas().drop(columns=['estudio','tipo','subtipo']), id_vars=['ID', 'repe', 'time'], var_name='axis')
        .set_index(["ID", "n_var", "time"])
        .to_xarray()
        .to_array()
        .squeeze("variable")
        .drop_vars("variable")
    )

    # Rename columns
    # dfAll = dfAll.rename({'abs time (s)':'time', 'Fx':'x', 'Fy':'y', 'Fz':'z'})

    # Assign extra coordinates
    daAll = assign_subcategories_xr(da=daAll, n_project=n_project)
    # daAll = daAll.assign_coords(estudio=('ID', dfAll.filter(pl.col('time')==0.000).get_column('estudio').to_list()),
    #                                          particip=('ID', dfAll.filter(pl.col('time')==0.000).get_column('particip').to_list()),
    #                                          tipo=('ID', dfAll.filter(pl.col('time')==0.000).get_column('tipo').to_list()),
    #                                          subtipo=('ID', dfAll.filter(pl.col('time')==0.000).get_column('subtipo').to_list()),
    #                                          repe=('ID', dfAll.filter(pl.col('time')==0.000).get_column('repe').to_list()),
    #                                          )
    # Set time coordinate type, required??
    ###########daAllFiles = daAllFiles.assign_coords(time=('time', daAllFiles.time.astype('float32').values))

    # daAllFiles.sel(ID='PCF_SCT05', axis='z').plot.line(x='time', col='repe')
    # daAllFiles.assign_coords(time=daAllFiles.time.astype('float32'))
    daAll.attrs = {
        "freq": (np.round(1 / (daAll.time[1] - daAll.time[0]), 1)).data,
        "units": "N",
    }
    daAll.time.attrs["units"] = "s"
    daAll.name = "Forces"

    return daAll


def read_isen_csv(
    file: str | Path,
    n_vars_load: List[str] | None = None,
    # to_dataarray: bool = False,
    engine="polars",
    raw: bool = False,
) -> Any:

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    if engine == "polars":
        da = read_isen_pl(
            file,
            n_vars_load=n_vars_load,
            raw=raw,
        )

    elif engine == "pandas":
        da = read_isen_pd(
            file,
            n_vars_load=n_vars_load,
            raw=raw,
        )

    else:
        raise ValueError(f"Engine {engine} not valid\nTry with 'polars' or 'pandas'")

    return da


def read_isen_pd(
    file: str | Path, n_vars_load: List[str] | None = None, raw: bool = False
) -> pd.DataFrame:
    # file = Path(r'F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\Piloto0iSen\S00_Todos-CaderasRodillas_00.csv')

    try:
        import pandas as pd
    except:
        raise ImportError("Pandas package not instaled")

    df = pd.read_csv(file, dtype=np.float64, engine="c")  # .astype(np.float64)
    # df = df.drop(columns=df.filter(regex="(Normal)") + df.filter(regex="(Tiempo.)"))

    #       df.filter(regex="(Normal)").values)).drop(columns=df.filter(regex="(Tiempo.)"))
    # df.filter(regex="(Normal)")+df.filter(regex="(Tiempo.)")
    # df.dropna(axis="columns", how="all")

    # df.filter(regex="Tiempo.")

    if raw:
        return df

    else:
        da = xr.DataArray()
        print("To dataarray not implemented yet")
        return da


def read_isen_pl(
    file: str | Path, n_vars_load: List[str] | None = None, raw: bool = False
) -> xr.DataArray | Any:
    # file = Path(r'F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\Piloto0iSen\S00_Todos-CaderasRodillas_00.csv')
    # file = Path(r'F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\Piloto0iSen\S00_Todos-Sensores_00.csv')

    try:
        import polars as pl
    except:
        raise ImportError(
            "Polars package not instaled. Install it if you want to use the accelerated version"
        )

    df = (
        pl.read_csv(
            file,
            # has_header=True,
            # skip_rows=0,
            # skip_rows_after_header=0,
            columns=n_vars_load,
            # separator=",",
        )  # , columns=nom_vars_cargar)
        .select(pl.exclude("^.*_duplicated_.*$"))  # quita columnas de tiempo duplicadas
        .select(pl.exclude(pl.String))  # quita columnas de texto con datos (Normal)
        # .with_columns(pl.all().cast(pl.Float64()))
    )

    """
    cadera = df.select(
            pl.col("^*Flexo-extensión cadera .*::Y$")
        )  # .to_numpy()
    """
    # df = df.to_pandas()

    if raw:
        return df
    # ----Transform polars to xarray
    else:

        # Separa ejes articulares
        x = df.select(pl.col("^Flexo-extensión.*$")).to_numpy()
        y = df.select(pl.col("^Aducción-abducción.*$")).to_numpy()
        z = df.select(pl.col("^Rotación.*$")).to_numpy()
        data = np.stack([x, y, z])
        freq = 1 / (df[1, "abs time (s)"] - df[0, "abs time (s)"])
        ending = -3
        coords = {
            "axis": ["x", "y", "z"],
            "time": np.arange(data.shape[1]) / freq,
            "n_var": ["Force"],  # [x[:ending] for x in df.columns if 'x' in x[-1]],
        }
        da = (
            xr.DataArray(
                data=data,
                dims=coords.keys(),
                coords=coords,
            )
            .astype(float)
            .transpose("n_var", "axis", "time")
        )
        da.name = "Forces"
        da.attrs["freq"] = freq
        da.time.attrs["units"] = "s"
        da.attrs["units"] = "N"

        return da


def load_merge_iSen_angles_csv(
    path: str | Path | None = None,
    file_list: List[str | Path] | None = None,
    n_vars_load: List[str] | None = None,
    n_project: str | None = None,
    data_type: str | None = None,
    show: bool = False,
) -> xr.DataArray:
    """
    Parameters
    ----------
    path : TYPE
        DESCRIPTION.
    file_list : TYPE, optional
        DESCRIPTION. List of files to read, overrides path.The default is None.
    n_vars_load : TYPE, optional
        DESCRIPTION. The default is None.
    n_project : string, optional
        DESCRIPTION. The name of the study.
    data_type:
        Conversión al tipo de datos indicado. Por defecto es None, que quiere
        decir que se mantiene el tipo original ('float64')
    show : bool, optional
        DESCRIPTION. The default is False.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    daAllFiles : xarray DataArray
        DESCRIPTION.
    dfAllFiles : pandas DataFrame
        DESCRIPTION.

    """
    try:
        import polars as pl
    except:
        raise ImportError(
            "Polars package not instaled. Install it if you want to use the accelerated version"
        )

    if data_type is None:
        data_type = float

    if file_list is None:
        file_list = sorted(
            list(path.glob("*.csv"))  # "**/*.csv"
        )  #'**/*.txt' incluye los que haya en subcarpetas
        file_list = [
            x
            for x in file_list
            if "error" not in x.name
            and "Sensores" not in x.name
            and "Actual" not in x.name
        ]  # selecciona archivos
    else:
        file_list = file_list

    # file = Path("F:/Investigacion/Proyectos/Tesis/TesisCoralPodologa/Registros/PRUEBASiSEN/ANGULOS.csv")

    print("\nLoading files...")
    timer_load = time.perf_counter()  # inicia el contador de tiempo

    n_processed_files = 0
    dfAll = (
        []
    )  # guarda los dataframes que va leyendo en formato lista y al final los concatena
    daAll = []
    error_files = []  # guarda los nombres de archivo que no se pueden abrir y su error
    for nf, file in enumerate(file_list):
        print(f"Loading file num. {nf+1}/{len(file_list)}: {file.name}")
        try:
            timerSub = time.perf_counter()  # inicia el contador de tiempo

            dfProvis = read_isen_pl(file, n_vars_load, raw=True)

            if len(file.stem.split("_")) == 5:
                n_project = file.stem.split("_")[0] if n_project is None else n_project
                particip = file.stem.split("_")[-4]
                tipo = file.stem.split("_")[-3]
                subtipo = file.stem.split("_")[-2]
            elif len(file.stem.split("_")) == 4:
                # estudio = file.stem.split("_")[0]
                particip = file.stem.split("_")[0]
                tipo = file.stem.split("_")[-3]
                subtipo = file.stem.split("_")[-2]
            elif len(file.stem.split("_")) == 3:
                particip = file.stem.split("_")[0]
                tipo = file.stem.split("_")[-2]
                subtipo = "X"
            elif len(file.stem.split("_")) == 2:
                particip = "X"
                tipo = "X"
                subtipo = "X"
            if n_project is None:
                n_project = "projectX"

            try:
                repe = str(
                    int(file.stem.split("_")[-1])
                )  # int(file.stem.split('.')[0][-1]
            except:
                repe = 0
            ID = f"{particip}_{tipo}_{subtipo}_{repe}"  # f'{n_project}_{file.stem.split("_")[0]}_{tipo}_{subtipo}'

            # freq = np.round(1/(dfProvis['time'][1]-dfProvis['time'][0]),1)

            # Añade categorías
            dfProvis = dfProvis.with_columns(
                [
                    pl.lit(n_project).alias("estudio"),
                    pl.lit(tipo).alias("tipo"),
                    pl.lit(subtipo).alias("subtipo"),
                    pl.lit(ID).alias("ID"),
                    pl.lit(particip).alias("particip"),
                    pl.lit(repe).alias("repe"),
                ]
            )  # .select(['estudio', 'tipo', 'subtipo', 'ID', 'repe'] + nom_vars_cargar)

            dfAll.append(dfProvis)

            print(f"{dfAll[-1].shape[0]} filas y {dfAll[-1].shape[1]} columnas")
            print("Tiempo {0:.3f} s \n".format(time.perf_counter() - timerSub))
            n_processed_files += 1

        except Exception as err:  # Si falla anota un error y continúa
            print(
                "\nATTENTION. Unable to process {0}, {1}, {2}".format(
                    file.parent.name, file.name, err
                ),
                "\n",
            )
            error_files.append(file.parent.name + " " + file.name + " " + str(err))
            continue

    dfAll = pl.concat(dfAll)

    print(
        f"Loaded {n_processed_files} files in {time.perf_counter()-timer_load:.3f} s \n"
    )

    # Si no ha podido cargar algún archivo, lo indica
    if len(error_files) > 0:
        print("\nATTENTION. Unable to load:")
        for x in range(len(error_files)):
            print(error_files[x])

    if isinstance(data_type, str):
        cast = pl.Float32() if data_type == "float32" else pl.Float64()
        dfAll = dfAll.select(
            # pl.col(['n_project', 'tipo', 'subtipo', 'ID', 'repe']),
            pl.exclude(n_vars_load),
            pl.col(n_vars_load).cast(cast),
        )

    dfAll = dfAll.rename({"Tiempo": "time"})  # , 'Fx':'x', 'Fy':'y', 'Fz':'z'})

    daAll = df_to_da(dfAll, n_project=n_project)

    daAll = split_dim_axis(daAll)

    daAll = split_dim_side(daAll)

    # TODO: CONTINUE TO MAKE IT SO THAT IF IT LOADS VARIABLES OTHER THAN THE CONVENTIONAL ONES, IT PASSES THEM TO DAARRAY BUT WITHOUT AXES
    # if  dfAll.columns == ['abs time (s)', 'Fx', 'Fy', 'Fz'] or n_vars_load == ['abs time (s)', 'Fx', 'Fy', 'Fz', 'Fx_duplicated_0', 'Fy_duplicated_0', 'Fz_duplicated_0']:
    #     daAll = pasa_dfpl_a_da(dfAll, merge_2_plats=merge_2_plats, show=show)

    """    
    #Transforma a pandas y a dataarray
    daAll = (dfAll.drop(['estudio','tipo','subtipo'])
                .melt(id_vars=['ID', 'repe', 'time'], variable_name='axis')
                .to_pandas()
                .set_index(['ID','repe', 'axis', 'time'])
                .to_xarray().to_array()
                .squeeze('variable').drop_vars('variable')
                )
     
    
    #Si hay 2 plataformas las agrupa en una
    if merge_2_plats==0:
        plat1 = (daAll.sel(axis=['Fx', 'Fy', 'Fz'])
                 .assign_coords(axis=['x', 'y', 'z'])
                 )
        plat2 = (daAll.sel(axis=['Fx_duplicated_0', 'Fy_duplicated_0', 'Fz_duplicated_0'])
                 .assign_coords(axis=['x', 'y', 'z'])
                 )
        daAll = (xr.concat([plat1, plat2], dim='plat')
                    .assign_coords(plat = [1, 2])
                    .transpose('ID', 'plat', 'repe', 'axis', 'time')
                    )
                             
    elif merge_2_plats==1:
        daAll = (daAll.sel(axis=['Fx', 'Fy', 'Fz'])
                     .assign_coords(axis=['x', 'y', 'z'])
                     )
    
    elif merge_2_plats==2:
        daAll = (daAll.sel(axis=['Fx_duplicated_0', 'Fy_duplicated_0', 'Fz_duplicated_0'])
                     .assign_coords(axis=['x', 'y', 'z'])
                     )
        
                    
    elif merge_2_plats==3:
       plat1 = (daAll.sel(axis=['Fx', 'Fy', 'Fz'])
                .assign_coords(axis=['x', 'y', 'z'])
                )
       plat2 = (daAll.sel(axis=['Fx_duplicated_0', 'Fy_duplicated_0', 'Fz_duplicated_0'])
                .assign_coords(axis=['x', 'y', 'z'])
                )
       daAll = plat1 + plat2
    
    daAll = asigna_subcategorias_xr(da=daAll, n_project=estudio, subtipo='2PAP')
    
    daAll.name = 'Forces'
    daAll.attrs['units'] = 'N'
    daAll.attrs['freq'] = freq
    daAll.time.attrs['units'] = 's'
    """
    return daAll


def load_merge_iSen_sensors_csv(
    path: str | Path | None = None,
    file_list: List[str | Path] | None = None,
    n_vars_load: List[str] | None = None,
    n_project: str | None = None,
    data_type: str | None = None,
    show: bool = False,
) -> xr.DataArray:
    """
    Load data from iSen CSV files and merge them into a single xarray DataArray.

    Parameters
    ----------
    path : str or Path
        Path to the folder containing the iSen CSV files.
    n_vars_load : None or list of str, optional
        List of variables to load from the CSV files. If None, all variables
        are loaded. The default is None.
    n_project : str, optional
        Name of the study. The default is None.
    data_type : type, optional
        Data type to which the values will be converted. If None, the original
        data type is kept. The default is None.
    show : bool, optional
        If True, the resulting DataArray is plotted. The default is False.

    Returns
    -------
    daAll : xarray DataArray
        DataArray containing all the data from the CSV files.

    """

    try:
        import polars as pl
    except:
        raise ImportError(
            "Polars package not instaled. Install it if you want to use the accelerated version"
        )

    if data_type is None:
        data_type = float

    if file_list is None:
        file_list = sorted(
            list(path.glob("*.csv"))  # "**/*.csv"
        )  #'**/*.txt' incluye los que haya en subcarpetas
        file_list = [
            x
            for x in file_list
            if "error" not in x.name
            and "aceleración local" not in x.name
            and "Actual" not in x.name
        ]  # selecciona archivos
    else:
        file_list = file_list

    # file = Path("F:/Investigacion/Proyectos/Tesis/TesisCoralPodologa/Registros/PRUEBASiSEN/ANGULOS.csv")

    print("\nLoading files...")
    timer_load = time.perf_counter()  # inicia el contador de tiempo

    num_processed_files = 0
    dfAll = (
        []
    )  # guarda los dataframes que va leyendo en formato lista y al final los concatena
    daAll = []
    error_files = []  # guarda los nombres de archivo que no se pueden abrir y su error
    for nf, file in enumerate(file_list):
        print(f"Loading file num. {nf+1}/{len(file_list)}: {file.name}")
        try:
            timerSub = time.perf_counter()  # inicia el contador de tiempo

            dfProvis = read_isen_pl(file, n_vars_load, raw=True)

            # ADAPT DIRECTORY LABELS
            if len(file.stem.split("_")) == 5:
                n_project = file.stem.split("_")[0] if n_project is None else n_project
                particip = file.stem.split("_")[-4]
                tipo = file.stem.split("_")[-3]
                subtipo = file.stem.split("_")[-2]
            elif len(file.stem.split("_")) == 4:
                # estudio = file.stem.split("_")[0]
                particip = file.stem.split("_")[0]
                tipo = file.stem.split("_")[-3]
                subtipo = file.stem.split("_")[-2]
            elif len(file.stem.split("_")) == 3:
                particip = file.stem.split("_")[0]
                tipo = file.stem.split("_")[-2]
                subtipo = "X"
            else:
                particip = file.stem.split("_")[0]
                tipo = "X"
                subtipo = "X"
                repe = 0

            if n_project is None:
                n_project = "EstudioX"

            try:
                repe = str(
                    int(file.stem.split("_")[-1])
                )  # int(file.stem.split('.')[0][-1]
            except:
                repe = 0
            ID = f"{particip}_{tipo}_{subtipo}_{repe}"  # f'{n_project}_{file.stem.split("_")[0]}_{tipo}_{subtipo}'

            # freq = np.round(1/(dfProvis['time'][1]-dfProvis['time'][0]),1)

            # Añade categorías
            dfProvis = dfProvis.with_columns(
                [
                    pl.lit(n_project).alias("estudio"),
                    pl.lit(tipo).alias("tipo"),
                    pl.lit(subtipo).alias("subtipo"),
                    pl.lit(ID).alias("ID"),
                    pl.lit(particip).alias("particip"),
                    pl.lit(repe).alias("repe"),
                ]
            )  # .select(['estudio', 'tipo', 'subtipo', 'ID', 'repe'] + nom_vars_cargar)

            dfAll.append(dfProvis)

            print(f"{dfAll[-1].shape[0]} rows and {dfAll[-1].shape[1]} columns")
            print("Time {0:.3f} s \n".format(time.perf_counter() - timerSub))
            num_processed_files += 1

        except Exception as err:  # Si falla anota un error y continúa
            print(
                "\nATTENTION. Unable to process {0}, {1}, {2}".format(
                    file.parent.name, file.name, err
                ),
                "\n",
            )
            error_files.append(file.parent.name + " " + file.name + " " + str(err))
            continue

    dfAll = pl.concat(dfAll)

    print(
        f"Loaded {num_processed_files} files in {time.perf_counter()-timer_load:.3f} s \n"
    )

    # Si no ha podido cargar algún archivo, lo indica
    if len(error_files) > 0:
        print("\nATTENTION. Unable to load:")
        for x in range(len(error_files)):
            print(error_files[x])

    if isinstance(
        data_type, str
    ):  # si se ha definido algún tipo de datos, por defecto es 'float64'
        cast = pl.Float32() if data_type == "float32" else pl.Float64()
        dfAll = dfAll.select(
            # pl.col(['n_project', 'tipo', 'subtipo', 'ID', 'repe']),
            pl.exclude(n_vars_load),
            pl.col(n_vars_load).cast(cast),
        )

    dfAll = dfAll.rename({"Tiempo": "time"})  # , 'Fx':'x', 'Fy':'y', 'Fz':'z'})

    daAll = df_to_da(dfAll, n_project=n_project)

    return daAll


def load_merge_iSen_csvxxxx(path, n_preprocessed_file):
    """TESTING"""
    print("\nLoading  iSen files...")
    timer_load = time.time()  # inicia el contador de tiempo
    # nomVarsACargar = nomVarsContinuas#+nomVarsDiscretas

    # Selecciona los archivos en la carpeta
    file_list = sorted(
        path.glob("*.csv")  # "**/*.csv"
    )  # list((path).glob('**/*.csv'))#incluye los que haya en subcarpetas
    file_list = [
        x for x in file_list if "error" not in x.name and "SENSORES" not in x.name
    ]  # selecciona archivos

    print("\nLoading files...")
    timer_load = time.perf_counter()  # inicia el contador de tiempo

    num_processed_files = 0
    dfAll = (
        []
    )  # guarda los dataframes que va leyendo en formato lista y al final los concatena
    error_files = []  # guarda los nombres de archivo que no se pueden abrir y su error

    for nf, file in enumerate(file_list[:]):
        print(f"Loading file num. {nf}/{len(file_list)}: {file.name}")
        """
        #Selecciona archivos según potencia
        if file.parent.parts[-1] not in ['2W', '6W']: #si no es de marcha se lo salta
            #print('{} archivo estático'.format(file.stem))
            continue
        """
        # # Selecciona archivos según variable
        # if file.stem not in [
        #     r"Flexo-extensión caderas",
        #     r"Flexo-extensión rodillas",
        #     r"Dorsiflexión-plantarflexión tobillos",
        # ]:  # si no es de marcha se lo salta
        #     # print('{} archivo estático'.format(file.stem))
        #     continue
        # print(file)

        try:
            timerSub = time.time()  # inicia el contador de tiempo
            dfprovis = pd.read_csv(
                file, sep=","
            )  # .filter(regex="::Y")  # se queda solo con las columnas de datos, descarta las de tiempos

            # Ajusta los nombres R y L
            dfprovis.rename(
                columns={
                    str(dfprovis.filter(regex="derech").columns.values)[2:-2]: "R",
                    str(dfprovis.filter(regex="izquierd").columns.values)[2:-2]: "L",
                },
                inplace=True,
            )
            # dfprovis.columns=['R', 'L']

            freq = (
                1 / dfprovis["Tiempo"].diff().mean()
            )  # (dfprovis["Tiempo"][1] - dfprovis["Tiempo"][0])

            # Añade columna tiempo
            t = np.arange(0, len(dfprovis) / freq)[
                : len(dfprovis)
            ]  # ajuste al cargar el tiempo porque en algunos t sale con un dato de más

            nom = file.parent.parts[-2]
            pot = file.parent.parts[-1]
            artic = file.stem.split()[-1][:-1]
            if "flex" in file.stem.lower():
                eje = "x"

            """
            if file.stem == 'Dorsiflexión-plantarflexión tobillos':
                artic = 'tobillo'
                eje='x'
            elif file.stem == 'Flexo-extensión rodillas':
                artic = 'rodilla'
                eje='x'
            elif file.stem == 'Flexo-extensión caderas':
                artic = 'cadera'
                eje='x'
            """

            # Añade etiquetas
            dfprovis = dfprovis.assign(
                **{
                    "nombre": nom,
                    "potencia": pot,
                    "articulacion": artic,
                    "eje": eje,
                    "time": t,
                }
            ).reset_index(drop=True)

            # Para formato tidy
            dfprovis = dfprovis.reindex(
                columns=["nombre", "potencia", "articulacion", "eje", "time", "R", "L"]
            )

            # Transforma a formato long, mejor dejarlo en tidy para que ocupe menos al guardarlo
            # dfprovis = pd.melt(dfprovis, id_vars=['nombre', 'potencia', 'articulacion', 'eje', 'time'], var_name='lado')#, value_vars=dfprovis[pot].iloc[:, :-4])
            # dfprovis = dfprovis.reindex(columns=['nombre', 'potencia', 'articulacion', 'eje', 'lado', 'time', 'value'])

            dfAllFiles.append(dfprovis)

            print(
                "{0} filas y {1} columnas".format(
                    dfAllFiles[-1].shape[0], dfAllFiles[-1].shape[1]
                )
            )
            print("Tiempo {0:.3f} s \n".format(time.time() - timerSub))
            num_processed_files += 1

        except Exception as err:  # Si falla anota un error y continua
            print(
                "\nATTENTION. Unable to process {0}, {1}, {2}".format(
                    file.parent.name, file.name, err
                ),
                "\n",
            )
            error_files.append(file.parent.name + " " + file.name + " " + str(err))
            continue
    dfAllFiles = pd.concat(dfAllFiles)
    print(f"Loaded {num_processed_files} files in {time.time() - timer_load:.3f} s \n")

    # Si no ha podido cargar algún archivo, lo indica
    if len(error_files) > 0:
        print("\nATTENTION. Unable to load:")
        for x in range(len(error_files)):
            print(error_files[x])

    # =============================================================================
    # Lo pasa a DataArray
    # =============================================================================
    # Transforma a formato long
    dfAllMulti = (
        pd.melt(
            dfAllFiles,
            id_vars=["nombre", "potencia", "articulacion", "eje", "time"],
            var_name="lado",
        )
        .reindex(
            columns=[
                "nombre",
                "potencia",
                "articulacion",
                "eje",
                "lado",
                "time",
                "value",
            ]
        )
        .set_index(["nombre", "potencia", "articulacion", "eje", "lado", "time"])
    )

    try:
        dfAllSubjects = dfAllMulti.to_xarray().to_array()
        dfAllSubjects = dfAllSubjects.sel(time=slice(0, 63))
        del dfAllSubjects["variable"]  # la quita de coordenadas
        dfAllSubjects = dfAllSubjects.squeeze("variable")  # la quita de dimensiones
        dfAllSubjects.attrs["freq"] = freq
        dfAllSubjects.attrs["units"] = "degrees"
        dfAllSubjects.time.attrs["units"] = "s"

        if False:
            dfAllSubjects.sel(
                nombre="Javi", potencia="2W", articulacion="rodilla", eje="x"
            ).plot.line(x="time", hue="lado")
            dfAllSubjects.sel(potencia="2W", articulacion="cadera").plot.line(
                x="time", row="nombre", col="lado"
            )

    except:
        print(
            "There may be a duplicate file. Search for the file with different duration."
        )
        # Si no funciona el data array, comprueba si hay duplicados
        for n, df in dfAllFiles.set_index(list(dfAllFiles.columns[:-2])).groupby(
            list(dfAllFiles.columns[:-2])
        ):
            print(n, len(df))

    if False:
        # Comparativa Todas las variables de side L y R juntas en una misma hoja de cada sujeto
        nompdf = work_path / "CamparacionVicon_iSen_PorArtics_iSen.pdf"
        with PdfPages(nompdf) as pdf_pages:
            for n, gda in dfAllSubjects.sel(eje="x").groupby("nombre"):
                g = gda.plot.line(
                    x="time",
                    row="potencia",
                    col="articulacion",
                    hue="lado",
                    sharey=False,
                    aspect=1.5,
                )
                for h, ax in enumerate(g.axes):  # extrae cada fila
                    for i in range(len(ax)):  # extrae cada axis (gráfica)
                        nom = str(g.data.loc[g.name_dicts[h, i]].nombre.data)
                        pot = str(g.data.loc[g.name_dicts[h, i]].potencia.data)
                        # print(nom, pot)
                        try:
                            ax[i].axvline(
                                x=dfFrames.loc[(nom, pot), "ini"] / freq, c="r", ls="--"
                            )
                            ax[i].axvline(
                                x=dfFrames.loc[(nom, pot), "fin"] / freq, c="r", ls="--"
                            )
                        except:
                            continue

                g.fig.subplots_adjust(top=0.95)
                g.fig.suptitle(n)
                pdf_pages.savefig(g.fig)
                plt.show()

    # =============================================================================
    #   Guarda archivos cargados
    # =============================================================================
    # Guarda xarray
    tpoGuardar = time.time()
    dfAllSubjects.to_netcdf(
        (work_path / (n_preprocessed_file + "_iSen")).with_suffix(".nc")
    )
    print(
        "\nPreprocessed Dataframe saved {0} en {1:0.2f} s.".format(
            n_preprocessed_file + "_iSen.nc", time.time() - tpoGuardar
        )
    )

    # Guarda dataframetpoGuardar = time.time()
    tpoGuardar = time.time()
    dfAllFiles.to_csv(
        (work_path / (n_preprocessed_file + "_iSen_tidy")).with_suffix(".csv"),
        index=False,
    )
    print(
        "\nPreprocessed DataArray saved {0} en {1:0.2f} s.".format(
            n_preprocessed_file + "_iSen_tidy.csv", time.time() - tpoGuardar
        )
    )
    # Transforma a formato long
    # dfAllFiles = pd.melt(dfAllFiles, id_vars=['nombre', 'potencia', 'articulacion', 'eje', 'time'], var_name='lado')#, value_vars=dfprovis[pot].iloc[:, :-4])
    # dfAllFiles = dfAllFiles.reindex(columns=['nombre', 'potencia', 'articulacion', 'eje', 'lado', 'time', 'value'])


def load_preprocessed(work_path: str | Path, n_preprocessed_file: str):
    # LOAD VICON
    if Path(
        (work_path / (n_preprocessed_file + "_Vicon")).with_suffix(".nc")
    ).is_file():
        tpo = time.time()
        ds_Vicon = xr.load_dataset(
            (work_path / (n_preprocessed_file + "_Vicon")).with_suffix(".nc")
        )
        # daAllFiles = xr.load_dataarray((work_path / (n_preprocessed_file+'_Vicon')).with_suffix('.nc'))
        print(
            "\nLoading preprocessed file ",
            n_preprocessed_file + "_Vicon.nc en {0:.3f} s.".format(time.time() - tpo),
        )
    else:
        raise Exception("Preprocessed Vicon file not found.")

    # LOAD ISEN
    if Path((work_path / (n_preprocessed_file + "_iSen")).with_suffix(".nc")).is_file():
        tpo = time.time()
        da_iSen = xr.load_dataarray(
            (work_path / (n_preprocessed_file + "_iSen")).with_suffix(".nc")
        )
        print(
            "\nSaved preprocessed file ",
            (work_path / (n_preprocessed_file + "_iSen")).with_suffix(".nc").name,
            "en {0:.3f} s.".format(time.time() - tpo),
        )
    else:
        raise Exception("Preprocessed iSen file not found")

    return ds_Vicon, da_iSen


# =============================================================================
# %% TESTS
# =============================================================================
if __name__ == "__main__":

    # Vicon
    # work_path = Path('F:\Programacion\Python\Mios\TratamientoDatos\EjemploBikefitting')
    work_path = Path(r"src\datasets")
    # dfAllFiles, daAllFiles = carga_preprocesa_vicon(path=work_path)
    file = work_path / "iSen_angles.csv"
    daData = read_isen_csv(file, engine="polars", raw=True)

    file = work_path / "iSen_angles.csv"
    daData = load_merge_iSen_angles_csv(file_list=[file])
    daData.sel(axis="x", time=slice(30, None)).plot.line(
        x="time", col="n_var", row="ID"
    )

    file = work_path / "iSen_sensors.csv"
    daData = load_merge_iSen_sensors_csv(file_list=[file])
    daData.sel(axis="x", time=slice(30, None)).plot.line(
        x="time", col="n_var", row="ID"
    )

    n_vars_load = (
        None  # ['Tiempo', 'BOLTWOODITE X::Y', 'BOLTWOODITE Y::Y', 'BOLTWOODITE Z::Y',
    )
    #                'LEO X::Y', 'LEO Y::Y', 'LEO Z::Y',
    #                'GASPEITE X::Y', 'GASPEITE Y::Y', 'GASPEITE Z::Y',
    #                'MIARGYRITE X::Y', 'MIARGYRITE Y::Y', 'MIARGYRITE Z::Y',
    #                'KARENAI X::Y', 'KARENAI Y::Y', 'KARENAI Z::Y'
    # ]
    daSensors = load_merge_iSen_sensors_csv(
        path=work_path / "SensoresSeparado", n_vars_load=n_vars_load
    )
    daSensors.sel(time=slice(30, None)).isel(ID=0).plot.line(
        x="time", col="n_var", col_wrap=4
    )

    r"""
    file = Path(r'F:\Investigacion\Proyectos\Tesis\TesisCoralPodologa\Registros\Piloto0iSen\S00_Todos-CaderasRodillas_00.csv')
    #Sigue siendo más rápido con polars que pd engine='c'
    t = time.perf_counter()  # inicia el contador de tiempo
    for i in range(100):
        dfProvis = read_isen_pd(file)
    print(time.perf_counter()-t)
    
    
    t = time.perf_counter()  # inicia el contador de tiempo
    for i in range(100):
        dfProvis = read_isen_pl(file)
    print(time.perf_counter()-t)
    """



# %% -*- coding: utf-8 -*-
"""
Functions to perform DLT operations with markers from video.
Based on xarray.

@author: Jose L. L. Elvira
"""


# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

import numpy as np
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt

try:
    from dltx import dlt_calibrate, dlt_reconstruct
except ImportError:
    raise ImportError(
        "dltx is not installed. Please install it with 'pip install dltx'"
    )
# --------------------------------------------------------------


__author__ = "Jose L. L. Elvira"
__version__ = "1.0.0"
__date__ = "15/02/2026"


"""
Updates:
    15/02/2026, v1.0.0
        - Initial release. 

    
"""


# =============================================================================
# %% FUNCTIONS
# =============================================================================


def dlt_calib(
    xyz: pd.DataFrame | np.ndarray,
    uv: list | pd.DataFrame | np.ndarray,
    show: bool = True,
) -> tuple[np.ndarray, float]:
    """
    Camera calibration by DLT using known object points and their image points.
    At least 6 calibration points for the 3D DLT and 4
    for the 2D DLT.

    Parameters
    ----------
    xyz : pd.DataFrame | np.ndarray
        Coordinates in the object space, num_points as rows and num_dims as columns (x, y) or (x, y, z).
    uv : list | pd.DataFrame | np.ndarray
        Coordinates in the image space, num_points as rows and num_dims as columns (u, v).
        If uv is a list, it must contain the uv coordinates for each camera.

    Returns
    -------
    L : np.ndarray
        Calibration parameters (8,) or (11,).
    err_image : float
        RMS error of the DLT transformation in units of image space.
    err_real : float
        RMS error of the DLT transformation in units of real object space.
    """
    xyz = np.asarray(xyz)
    uv = np.asarray(uv)
    err_real = np.nan  # only in 2D reconstruction
    if uv.ndim == 2:
        uv = np.expand_dims(uv, axis=0)

    num_dims = xyz.shape[1]
    num_cams = uv.shape[0]
    num_points = uv.shape[1]

    L = []
    err_image = []

    for i, _uv in enumerate(uv):
        _L, _err_image = dlt_calibrate(num_dims, xyz, _uv)
        L.append(_L)
        err_image.append(_err_image)
        print(f"Calibrated Camera {i}")
    L = np.array(L)
    err_image = np.array(err_image)

    # TODO: join as a general case? uv[0,i], uv[:,i]
    if num_cams == 1:
        # L, err_image = dlt_calibrate(num_dims, xyz, uv)

        # Reconstruct original points after calibration
        xyz_recons = np.zeros((num_points, num_dims))
        for i in range(num_points):
            try:
                xyz_recons[i, :] = dlt_reconstruct(num_dims, num_cams, L, uv[0, i])
            except np.linalg.LinAlgError:
                xyz_recons[i, :] = np.nan
            except Exception as e:
                print(f"Error in dlt_reconstruct for point {i}: {e}")
                xyz_recons[i, :] = np.nan

    else:
        xyz_recons = np.zeros((num_points, num_dims))
        for i in range(num_points):
            try:
                xyz_recons[i, :] = dlt_reconstruct(num_dims, num_cams, L, uv[:, i])
            except np.linalg.LinAlgError:
                xyz_recons[i, :] = np.nan

    err_real = np.mean(np.linalg.norm((xyz_recons - xyz), axis=1))
    # err_real = np.mean(np.sqrt(np.sum((np.array(xyz_recons) - np.array(xyz)) ** 2, 1)))

    if show:
        if num_dims == 2:
            fig, ax = plt.subplots()
            ax.plot(xyz[:, 0], xyz[:, 1], "bo", alpha=0.6, label="Real")
            ax.plot(
                xyz_recons[:, 0],
                xyz_recons[:, 1],
                "ro",
                alpha=0.6,
                label="Reconst",
            )
            # ax.text(
            #     0.5,
            #     0.95,
            #     f"Error original: {err_real:.4f} (original units)",
            #     transform=ax.transAxes,
            #     fontsize=8,
            #     c="r",
            #     horizontalalignment="center",
            # )
            # ax.text(
            #     0.5,
            #     0.90,
            #     f"Error image: {err_image[cam]:.4f} (image units)",
            #     transform=ax.transAxes,
            #     fontsize=8,
            #     c="r",
            #     horizontalalignment="center",
            # )
            ax.set_aspect("equal", "box")
            ax.set_title(
                "Real vs. reconstructed\n"
                f"Error orig: {err_real:.4f} (orig units)\n"
                f"Error img: {np.mean(err_image):.4f} (img units)"
            )
            ax.legend(fontsize=8)
            plt.show()

        elif num_dims == 3:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            ax.scatter(
                xyz[:, 0],
                xyz[:, 1],
                xyz[:, 2],
                c="b",
                marker="o",
                alpha=0.6,
                label="Real",
            )
            ax.scatter(
                xyz_recons[:, 0],
                xyz_recons[:, 1],
                xyz_recons[:, 2],
                c="r",
                marker="o",
                alpha=0.6,
                label="Reconst",
            )

            ax.set_title(
                "Real vs. reconstructed\n"
                f"Error orig: {err_real:.4f} (orig units)\n"
                f"Error img: {np.mean(err_image):.4f} (img units)"
            )
            ax.set_xlabel("X")
            ax.set_ylabel("Y")
            ax.set_zlabel("Z")
            ax.legend()
            plt.show()

    return L, err_image, err_real


def dlt_recons_xr(
    da_uv: xr.DataArray | list[xr.DataArray],
    L: np.ndarray,
    num_cams: int = 1,
    num_dims: int = 2,
) -> xr.DataArray:
    """
    Vectorized DLT reconstruction for 2D/3D case with N cameras using xarray.

    Parameters
    ----------
    da_uv : xr.DataArray or list of xr.DataArray
        DataArray with coordinates.
        If list: len(da_uv) must be equal to num_cams. Each DA must have 'axis' dim.
        If DataArray: must have 'camera' and 'axis' dimensions, or be a single camera (nc=1).
    L : np.array
        Calibration parameters.
        Shape (8,) or (9,) for 2D 1-cam; (11,) or (12,) for 3D 1-cam.
        Shape (nc, 8/9) or (nc, 11/12) for multi-cam.
    num_cams : int
        Number of cameras.
    num_dims : int
        Number of dimensions (2 or 3).

    Returns
    -------
    xr.DataArray
        Reconstructed coordinates. Shape (..., num_dims).
    """

    # Standardize input to single DataArray with 'camera' dimension
    if isinstance(da_uv, list):
        num_cams = len(da_uv)

        da_input = xr.concat(da_uv, dim="camera").assign_coords(
            {"camera": np.arange(num_cams)}
        )

    elif isinstance(da_uv, xr.DataArray):
        da_input = da_uv
        if num_cams == 1 and "camera" not in da_input.dims:
            # Ensure camera dimension exists for single camera
            da_input = da_input.expand_dims("camera")

        if num_cams > 1 and "camera" not in da_input.dims:
            raise ValueError(
                "For num_cams > 1, input DataArray must have 'camera' dimension."
            )
        num_cams = da_input.sizes["camera"]
    else:
        raise TypeError("da_uv must be a DataArray or a list of DataArrays.")

    # Standardize and validate L
    L = np.asarray(L)
    if num_cams == 1 and L.ndim == 1:
        L = L.reshape(1, -1)

    expected_params = {2: [8, 9], 3: [11, 12]}
    if num_dims not in expected_params or L.shape[1] not in expected_params[num_dims]:
        raise ValueError(
            f"L shape {L.shape} invalid for {num_dims}D. Expected ({num_cams}, {expected_params.get(num_dims, '?')})"
        )

    # Fast path for 2D single camera using direct matrix inversion
    Hinv = None
    if num_cams == 1 and num_dims == 2:
        try:
            L_flat = L.flatten()
            if L_flat.size == 8:
                L_flat = np.append(L_flat, 1.0)
            Hinv = np.linalg.inv(L_flat.reshape(3, 3))
        except Exception:
            pass  # Fallback to SVD

    def _recon_svd(uv_all, L_all):
        """Core reconstruction using SVD or fast path"""

        # Fast path for 2D 1-cam
        if Hinv is not None:
            uv = uv_all.squeeze(axis=-2)
            uv1 = np.concatenate([uv, np.ones(uv.shape[:-1] + (1,))], axis=-1)
            xyz = uv1 @ Hinv.T
            with np.errstate(divide="ignore", invalid="ignore"):
                return xyz[..., :2] / xyz[..., 2:3]

        # General SVD path
        input_shape = uv_all.shape
        batch_shape = input_shape[:-2]
        n_samples = int(np.prod(batch_shape)) if batch_shape else 1

        uv_flat = uv_all.reshape(n_samples, num_cams, 2)
        valid_mask = ~np.isnan(uv_flat).any(axis=(1, 2))

        X_out = np.full((n_samples, num_dims), np.nan)
        if not valid_mask.any():
            return X_out.reshape(batch_shape + (num_dims,))

        uv_valid = uv_flat[valid_mask]
        u, v = uv_valid[..., 0], uv_valid[..., 1]

        # Build M matrix - unified logic for 2D and 3D
        n_params = L_all.shape[1]
        L_norm = 1.0 if n_params in [8, 11] else L_all[:, -1]

        if num_dims == 3:
            # Extract 3D parameters
            L_vals = L_all[:, :11].T  # (11, nc)
            row1 = np.stack([L_vals[i] - u * L_vals[i + 7] for i in range(4)], axis=-1)
            row2 = np.stack(
                [L_vals[i + 3] - v * L_vals[i + 7] for i in range(4)], axis=-1
            )
            if n_params == 12:
                row1[..., -1] = L_vals[3] - u * L_norm
                row2[..., -1] = L_vals[7] - v * L_norm
        else:  # 2D
            L_vals = L_all[:, :8].T  # (8, nc)
            row1 = np.stack([L_vals[i] - u * L_vals[i + 5] for i in range(3)], axis=-1)
            row2 = np.stack(
                [L_vals[i + 2] - v * L_vals[i + 5] for i in range(3)], axis=-1
            )
            if n_params == 9:
                row1[..., -1] = L_vals[2] - u * L_norm
                row2[..., -1] = L_vals[5] - v * L_norm

        M = np.stack([row1, row2], axis=2).reshape(valid_mask.sum(), num_cams * 2, -1)

        # Solve using SVD
        try:
            _, _, Vh = np.linalg.svd(M)
            X = Vh[..., -1, :]
            with np.errstate(divide="ignore", invalid="ignore"):
                X_out[valid_mask] = X[..., :-1] / X[..., -1:]
        except np.linalg.LinAlgError:
            pass  # Return NaNs

        return X_out.reshape(batch_shape + (num_dims,))

    coord_axis = ["x", "y"] if num_dims == 2 else ["x", "y", "z"]

    return xr.apply_ufunc(
        _recon_svd,
        da_input.sel(axis=["x", "y"]),
        kwargs={"L_all": L},
        input_core_dims=[["camera", "axis"]],
        output_core_dims=[["axis"]],
        exclude_dims={"axis"},
        vectorize=False,
        dask="parallelized",
        output_dtypes=[float],
        dask_gufunc_kwargs={"output_sizes": {"axis": num_dims}},
    ).assign_coords({"axis": coord_axis})


# %% MAIN
if __name__ == "__main__":
    # Test with real data from numpy arrays
    # Calibration points in object space (centimetres)
    xy = np.array([[0, 0], [0, 45], [0, 90], [60, 0], [60, 45], [60, 90]])

    # Calibration points in image space (pixels)
    uv = np.array(
        [
            [195.99, 74.83],
            [193.43, 312.34],
            [188.3, 547.52],
            [512.51, 77.39],
            [510.06, 311.46],
            [508.88, 544.23],
        ]
    )

    # Calculate calibration parameters
    L, err_image, err_real = dlt_calib(xy, uv, show=True)

    # --------------------------------------
    # Test with simulated example prepared to read from csv files
    # Real object space dimensions (metres)
    xy = pd.DataFrame(
        columns=["x", "y"],
        data=[
            [0.0, 13.410],
            [6.1, 13.410],
            [6.1, 6.705],
            [6.1, 0.000],
            [0.0, 0.000],
            [0.0, 6.705],
        ],
    )
    # xy = pd.read_csv("xy.csv")

    # Digitised in image space dimensions, 3 frames of the same 6 points
    uv = pd.DataFrame(
        columns=["fot", "p", "x", "y"],
        data=[
            [0, 0, 743, 881],
            [0, 1, 1165, 894],
            [0, 2, 1230, 602],
            [0, 3, 1345, 99],
            [0, 4, 620, 76],
            [0, 5, 696, 587],
            [11, 0, 743, 881],
            [11, 1, 1165, 894],
            [11, 2, 1230, 601],
            [11, 3, 1344, 100],
            [11, 4, 620, 76],
            [11, 5, 697, 587],
            [12, 0, 744, 881],
            [12, 1, 1165, 893],
            [12, 2, 1230, 601],
            [12, 3, 1344, 100],
            [12, 4, 620, 76],
            [12, 5, 696, 587],
        ],
    )
    # uv = pd.read_csv("uv.csv")
    # Calculate the mean of each digitised point
    uv = uv.groupby("p").mean().drop(columns="fot")

    # calcula parámetros de calibración
    L, err_img, err_real = dlt_calib(xy, uv, show=True)

    # --------------------------------------
    # Test 3D with 4 cameras
    xyz = [
        [0, 0, 0],
        [0, 12.3, 0],
        [14.5, 12.3, 0],
        [14.5, 0, 0],
        [0, 0, 14.5],
        [0, 12.3, 14.5],
        [14.5, 12.3, 14.5],
        [14.5, 0, 14.5],
    ]

    uv1 = [
        [1302, 1147],
        [1110, 976],
        [1411, 863],
        [1618, 1012],
        [1324, 812],
        [1127, 658],
        [1433, 564],
        [1645, 704],
    ]
    uv2 = [
        [1094, 1187],
        [1130, 956],
        [1514, 968],
        [1532, 1187],
        [1076, 854],
        [1109, 647],
        [1514, 659],
        [1523, 860],
    ]
    uv3 = [
        [1073, 866],
        [1319, 761],
        [1580, 896],
        [1352, 1016],
        [1064, 545],
        [1304, 449],
        [1568, 557],
        [1313, 668],
    ]
    uv4 = [
        [1205, 1511],
        [1193, 1142],
        [1601, 1121],
        [1631, 1487],
        [1157, 1550],
        [1139, 1124],
        [1628, 1100],
        [1661, 1520],
    ]

    # Calibrate
    L, err_image, err_real = dlt_calib(xyz, [uv1, uv2, uv3, uv4], show=True)

    # ------------------------------------
    # Test 2D reconstruction

    xy = pd.DataFrame(
        columns=["x", "y"],
        data=[
            [0, 0],
            [0, 45],
            [0, 90],
            [60, 0],
            [60, 45],
            [60, 90],
        ],
    )

    uv = pd.DataFrame(
        columns=["x", "y"],
        data=[
            [195.99, 74.83],
            [193.43, 312.34],
            [188.3, 547.52],
            [512.51, 77.39],
            [510.06, 311.46],
            [508.88, 544.23],
        ],
    )

    # Calibrate Camera
    L, err_image, err_real = dlt_calib(xy, uv, show=True)

    # Digitised markers
    markers_uv = pd.DataFrame(
        columns=["1x", "1y", "2x", "2y", "3x", "3y", "4x", "4y"],
        index=np.arange(10) / 50,
        data=[
            [195.99, 74.83, 188.3, 547.52, 512.51, 77.39, 508.88, 544.23],
            [199.99, 89.55, 192.7, 548.45, 510.11, 77.39, 508.88, 543.86],
            [207.99, 116.94, 201.5, 549.18, 505.31, 77.39, 508.88, 543.18],
            [219.99, 153.13, 214.7, 549.60, 498.11, 77.39, 508.88, 542.27],
            [235.99, 193.05, 232.3, 549.67, 488.51, 77.39, 508.88, 541.27],
            [255.99, 231.09, 254.3, 549.36, 476.51, 77.39, 508.88, 540.32],
            [279.99, 261.91, 280.7, 548.72, 462.11, 77.39, 508.88, 539.55],
            [307.99, 286.25, 311.5, 546.85, 445.31, 77.40, 508.88, 539.07],
            [339.99, 286.25, 346.7, 545.88, 426.11, 77.41, 508.88, 538.95],
            [375.99, 276.25, 386.3, 545.88, 404.51, 77.42, 508.88, 539.20],
        ],
    )

    # Convert to dataarray
    x = markers_uv.filter(regex="x")
    y = markers_uv.filter(regex="y")
    daMarkers_uv = xr.DataArray(
        data=np.stack([x, y], axis=1),
        coords={
            "time": markers_uv.index,
            "axis": ["x", "y"],
            "marker": range(x.shape[1]),
        },
        dims=["time", "axis", "marker"],
    ).transpose("marker", "axis", "time")

    # Reconstruct in object space
    daMarkers_xy = dlt_recons_xr(
        da_uv=daMarkers_uv, L=L, num_cams=1, num_dims=2
    ).transpose("marker", "axis", "time")

    # Show reconstruction
    ax = xy.plot.scatter(x="x", y="y")
    daMarkers_xy.to_dataset(dim="axis").plot.scatter(
        x="x", y="y", marker="o", c="r", alpha=0.5, ax=ax
    )
    ax.set_aspect("equal", "box")
